"""Xilinx ISE Synthesis flow"""
import logging
from typing import Mapping, Union

from ...tool import Tool
from ...utils import try_convert
from ..flow import FpgaSynthFlow

logger = logging.getLogger(__name__)

OptionValueType = Union[str, int, bool, float]
OptionsType = Mapping[str, OptionValueType]


class IseSynth(FpgaSynthFlow):
    """FPGA synthesis using Xilinx ISE"""

    class Settings(FpgaSynthFlow.Settings):
        # see https://www.xilinx.com/support/documentation/sw_manuals/xilinx14_7/devref.pdf
        synthesis_options: OptionsType = {
            "Optimization Effort": "High",
            "Optimization Goal": "Speed",
            "Keep Hierarchy": "No",
        }
        map_options: OptionsType = {
            "Global Optimization": "Speed",
            "Map Effort Level": "High",
            "Placer Effort Level": "High",
            "Perform Timing-Driven Packing and Placement": True,
            "Combinatorial Logic Optimization": True,
        }
        pnr_options: OptionsType = {
            "Place & Route Effort Level (Overall)": "High",
        }
        trace_options: OptionsType = {
            "Report Type": "Verbose Report",
        }

    def run(self) -> None:
        xcf_file = self.copy_from_template("constraints.xcf")
        ucf_file = self.copy_from_template("constraints.ucf")

        self.add_template_filter(
            "quote_str", lambda v: f'"{v}"' if isinstance(v, str) else v
        )

        script_path = self.copy_from_template(
            "ise_synth.tcl",
            xcf_file=xcf_file,
            ucf_file=ucf_file,
        )
        xtclsh = Tool("xtclsh")
        xtclsh.run(script_path)

    def parse_reports(self) -> bool:
        top = self.design.rtl.top
        assert top
        # self.parse_report_regex(self.design.name + ".twr", r'(?P<wns>\-?\d+')
        fail = not self.parse_report_regex(
            top + "_par.xrpt",
            r'stringID="PAR_SLICES" value="(?P<slice>\-?\d+)"',
            r'stringID="PAR_SLICE_REGISTERS" value="(?P<ff>\-?\d+)"',
            r'stringID="PAR_SLICE_LUTS" value="(?P<lut>\-?\d+)"',
        )
        fail |= not self.parse_report_regex(
            top + ".syr",
            r"Minimum period:\s+(?P<minimum_period>\-?\d+(?:\.\d+)?)ns\s+\(Maximum Frequency: (?P<maximum_frequency>\-?\d+(?:\.\d+)?)MHz\)",
            r"Slack:\s+(?P<wns>\-?\d+(?:\.\d+)?)ns",
        )
        if "wns" in self.results:
            wns = try_convert(self.results["wns"])
            fail |= not isinstance(wns, (float, int)) or wns < 0
        return not fail



import logging
import os
import platform
from abc import ABCMeta
from functools import cached_property
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Union

from varname import argname

from ...dataclass import Field, validator
from ...design import Design, DesignSource, Tuple012, VhdlSettings
from ...gtkwave import gen_gtkw
from ...tool import Docker, Tool
from ...utils import SDF, common_root
from ..flow import Flow, FlowSettingsError, SimFlow, SynthFlow

log = logging.getLogger(__name__)


def _get_wave_opt_signals(wave_opt_file, extra_top=None):
    signals = []
    with open(wave_opt_file, "r") as f:
        for line in f.read().splitlines():
            line = line.strip()
            if line.startswith("/"):
                sig = line[1:].split("/")
                if extra_top:
                    sig.insert(0, extra_top)
                signals.append(sig)
    root_group = common_root(signals)
    return signals, root_group


def setting_flag(variable: Any, assign=True) -> List[str]:
    """skip if none"""
    if variable is None:
        return []
    var_name = argname("variable")
    if not isinstance(variable, (list, tuple)):
        variable = [variable]
    flags = []
    for v in variable:
        if v:
            flag = "--" + (var_name.replace("_", "-"))
            if isinstance(v, bool):
                flags.append(flag)
            elif assign:
                flags.append(flag + "=" + str(v))
            else:
                flags += [flag, v]
    return flags


class GhdlTool(Tool):
    """GHDL VHDL simulation, synthesis, and linting tool: https://ghdl.readthedocs.io"""

    docker = Docker(image="hdlc/sim:osvb")
    executable = "ghdl"

    @cached_property
    def _info(self) -> Dict[str, str]:
        out = self.run_get_stdout(
            "--version",
        )
        lines = [line.strip() for line in out.splitlines()]
        if len(lines) < 3:
            return {}
        self._version = tuple(lines[0].split("."))

        return {
            "version": ".".join(self.version),
            "compiler": lines[1],
            "backend": lines[2],
        }


class Ghdl(Flow, metaclass=ABCMeta):
    """VHDL simulation using GHDL"""

    ghdl = GhdlTool()  # pyright: reportGeneralTypeIssues=none

    class Settings(Flow.Settings):
        analysis_flags: List[str] = []
        elab_flags: List[str] = ["--syn-binding"]
        warn_flags: List[str] = [
            "--warn-binding",
            "--warn-default-binding",
            "--warn-reserved",
            "--warn-library",
            "--warn-vital-generic",
            "--warn-shared",
            "--warn-runtime-error",
            "--warn-body",
            "--warn-specs",
            "--warn-unused",
            "--warn-static",
            "--warn-parenthesis",
        ]
        werror: bool = Field(
            False, description="warnings are always considered as errors"
        )
        elab_werror: bool = Field(
            False, description="During elaboration, warnings are considered as errors"
        )
        relaxed: bool = Field(
            True,
            description="Slightly relax some rules to be compatible with various other simulators or synthesizers.",
        )
        clean: bool = Field(False, description="Run 'clean' before elaboration")
        diagnostics: bool = Field(
            True, description="Enable both color and source line carret diagnostics."
        )
        work: Optional[str] = Field(
            None, description="Set the name of the WORK library"
        )
        expect_failure: bool = False

        def common_flags(self, vhdl: VhdlSettings) -> List[str]:
            cf: List[str] = []
            if vhdl.standard:
                cf.append(f"--std={vhdl.standard}")
            if vhdl.synopsys:
                cf.append("-fsynopsys")
            if self.work:
                cf.append(f"--work={self.work}")
            cf += [f"-P{p}" for p in self.lib_paths]
            return cf

        @staticmethod
        def generics_flags(generics: Optional[Dict[str, Any]]) -> List[str]:
            if not generics:
                return []

            def conv(v):
                if isinstance(v, bool):
                    return str(v).lower()
                return v

            return [f"-g{k}={conv(v)}" for k, v in generics.items()]

        def get_flags(self, vhdl: VhdlSettings, stage: str) -> List[str]:
            common = self.common_flags(vhdl)
            warn_flags: List[str] = self.warn_flags
            analysis_flags: List[str] = common + self.analysis_flags
            elab_flags: List[str] = common + self.elab_flags
            find_top_flags = common
            if stage == "common":
                return common
            if stage == "remove":
                return common
            if self.werror:
                warn_flags += ["--warn-error"]
            elif self.elab_werror:
                elab_flags += ["--warn-error"]
            if self.relaxed:
                analysis_flags.extend(["-frelaxed-rules", "-frelaxed", "--mb-comments"])
                elab_flags.extend(["-frelaxed"])
            if self.verbose:
                analysis_flags.append("-v")
                elab_flags.append("-v")
            if self.diagnostics:
                elab_flags.extend(
                    [
                        "-fcaret-diagnostics",
                        "-fcolor-diagnostics",
                        "-fdiagnostics-show-option",
                    ]
                )
            if self.work:
                elab_flags.append("-v")
            if self.expect_failure:
                elab_flags.append("--expect-failure")
            if vhdl.synopsys:
                analysis_flags.append("--ieee=synopsys")
            if stage in ("import", "analyze"):
                return analysis_flags + warn_flags
            if stage in ("make", "elaborate"):
                return elab_flags + warn_flags
            if stage in ("find-top"):
                return find_top_flags
            raise ValueError("unknown stage!")

    def elaborate(
        self, sources: List[DesignSource], top: Union[str, Tuple012], vhdl: VhdlSettings
    ) -> Tuple012:
        """returns top(s) as a list"""
        assert isinstance(self.settings, self.Settings)
        ss = self.settings
        if isinstance(top, str):
            top = (top,)
        steps = ["import", "make"]
        if ss.clean:
            steps.insert(0, "remove")
        if not top:
            # run find-top after import
            log.warning("added find-top to steps")
            steps.insert(steps.index("import") + 1, "find-top")
        for step in steps:
            args = ss.get_flags(vhdl, step)
            if isinstance(ss, SimFlow.Settings):
                args += ss.optimization_flags
            if step in ["import", "analyze"]:
                args += [str(s) for s in sources]
            elif step in ["make", "elaborate"]:
                if self.ghdl.info.get("backend", "").lower().startswith("llvm"):
                    if platform.system() == "Darwin" and platform.machine() == "arm64":
                        args += ["-Wl,-Wl,-no_compact_unwind"]
                args += list(top)
            if step == "find-top":
                out = self.ghdl.run_get_stdout(step, *args)
                top_list = out.strip().split()
                # clunky way of converting to tuple, just to be safe and also keep mypy happy
                top = (
                    ()
                    if len(top_list) == 0
                    else (top_list[0],)
                    if len(top_list) == 1
                    else (top_list[0], top_list[1])
                )
                if top:
                    log.info("find-top: top-module was set to %s", top)
                else:
                    log.warning("find-top: unable to determine the top-module")
            else:
                self.ghdl.run(step, *args)
        return top


class GhdlSynth(Ghdl, SynthFlow):
    """
    Convert a VHDL design using 'ghdl --synth'
     (Please take a look at 'Yosys' flow (or other synthesis flows) for general VHDL, Verilog, or mixed-language synthesis targeting FPGAs or ASICs)
    """

    class Settings(Ghdl.Settings, SynthFlow.Settings):
        vendor_library: Optional[str] = Field(
            None, description="Any unit from this library is a black box"
        )
        no_formal: bool = Field(
            True,
            description="Neither synthesize assert nor PSL. Required for yosys+nextpnr flow.",
        )
        no_assert_cover: bool = Field(
            False, description="Cover PSL assertion activation"
        )
        assert_assumes: bool = Field(
            False, description="Treat all PSL asserts like PSL assumes"
        )
        assume_asserts: bool = Field(
            False, description="Treat all PSL assumes like PSL asserts"
        )
        out: Optional[
            Literal["vhdl", "raw-vhdl", "verilog", "dot", "none", "raw", "dump"]
        ] = Field(None, description="Type of output to generate")
        out_file: Optional[str] = None

    def run(self) -> None:
        design = self.design
        assert isinstance(self.settings, self.Settings)
        ss = self.settings
        top = self.elaborate(design.rtl.sources, design.rtl.top, design.language.vhdl)
        args = self.synth_args(ss, design, one_shot_elab=False, top=top)
        self.ghdl.run_stdout_to_file("synth", *args, stdout=ss.out_file)

    @staticmethod
    def synth_args(
        ss: Settings, design: Design, one_shot_elab: bool = True, top: Tuple012 = ()
    ) -> List[str]:
        flags = ss.get_flags(design.language.vhdl, "elaborate")
        flags += setting_flag(ss.vendor_library)
        flags += setting_flag(ss.out)
        flags += setting_flag(ss.no_formal)
        flags += setting_flag(ss.no_assert_cover)
        flags += setting_flag(ss.assert_assumes)
        flags += setting_flag(ss.assume_asserts)

        flags.extend(ss.generics_flags(design.rtl.generics))
        if one_shot_elab:
            flags += [str(v) for v in design.rtl.sources if v.type == "vhdl"]
            flags.append("-e")
        if not top:
            top = (design.rtl.top,)
        flags.extend(list(top))
        return flags


# class GhdlLint(Ghdl, Flow):
#     """Lint VHDL sources using GHDL"""

#     class Settings(Ghdl.Settings):
#         pass


class GhdlSim(Ghdl, SimFlow):
    """Simulate a VHDL design using GHDL"""

    cocotb_sim_name = "ghdl"

    class Settings(Ghdl.Settings, SimFlow.Settings):
        run_flags: List[str] = []
        optimization_flags: List[str] = Field(
            ["-O3"], description="Simulation optimization flags"
        )
        asserts: Optional[Literal["disable", "disable-at-0"]] = None
        ieee_asserts: Optional[Literal["disable", "disable-at-0"]] = "disable-at-0"
        sdf: SDF = Field(
            SDF(),
            description="Do VITAL annotation using SDF files(s). A single string is interpreted as a MAX SDF file.",
        )
        wave: Optional[str] = Field(
            None, description="Write the waveforms into a GHDL Waveform (GHW) file."
        )
        read_wave_opt: Optional[str] = Field(
            None,
            description="Filter signals to be dumped to the wave file according to the wave option file provided.",
        )
        write_wave_opt: Optional[str] = Field(
            None,
            description="Creates a wave option file with all the signals of the design. Overwrites the file if it already exists.",
        )
        fst: Optional[str] = Field(
            None, description="Write the waveforms into an _fst_ file."
        )
        stop_delta: Optional[str] = Field(
            None,
            description="Stop the simulation after N delta cycles in the same current time.",
        )
        disp_tree: Optional[Literal["inst", "proc", "port"]] = Field(
            None,
            description="Display the design hierarchy as a tree of instantiated design entities. See GHDL documentation for more details.",
        )
        vpi: Union[None, str, List[str]] = Field(
            None,
            description="Load VPI library (or multiple libraries)",
        )
        # TODO workdir?

        @validator("wave", "fst", pre=True)
        def validate_wave(cls, value, field):  # pylint: disable=no-self-argument
            if value is not None:
                ext = (
                    ".ghw"
                    if field.name == "wave"
                    else ".fst"
                    if field.name == "fst"
                    else ""
                )
                if isinstance(value, bool):
                    value = "dump" + ext if value else None
                else:
                    if not isinstance(value, str):
                        value = str(value)
                    if not value.endswith(ext):
                        value += ext
            return value

    def run(self) -> None:
        design = self.design
        assert design.tb
        assert isinstance(self.settings, self.Settings)
        ss = self.settings
        cf = ss.common_flags(design.language.vhdl)
        run_flags = self.settings.run_flags
        sdf_root = ss.sdf.root if ss.sdf.root else design.tb.uut
        for delay_type, sdf_file in ss.sdf.delay_items():
            if sdf_file:
                assert sdf_root, "neither SDF root nor tb.uut are provided"
                run_flags.append(f"--sdf={delay_type}={sdf_root}={sdf_file}")

        def fp(s: Union[str, os.PathLike]) -> str:
            if not os.path.isabs(s):
                return str(self.run_path.relative_to(Path.cwd()) / s)
            return str(s)

        if ss.vcd:
            if ss.vcd.endswith((".gz", ".vcdgz")):
                run_flags.append(f"--vcdgz={ss.vcd}")
            else:
                run_flags.append(f"--vcd={ss.vcd}")
            log.warning("Dumping VCD to %s", fp(ss.vcd))
        if ss.fst:
            run_flags += setting_flag(ss.fst)
            log.warning("Dumping fst to %s", fp(ss.fst))
        if ss.wave:
            run_flags += setting_flag(ss.wave)
            log.warning("Dumping GHW to %s", fp(ss.wave))

        if ss.wave or ss.vcd or ss.fst:
            if not ss.read_wave_opt and not ss.write_wave_opt:
                ss.write_wave_opt = "wave.opt"

        if ss.write_wave_opt:
            p = Path(ss.write_wave_opt)
            if p.exists():
                log.warning("Deleting existing wave option file: %s", p)
                p.unlink()
        run_flags += setting_flag(ss.write_wave_opt)
        if ss.read_wave_opt:
            if not Path(ss.read_wave_opt).exists():  # TODO move to validation
                raise FlowSettingsError(
                    [
                        (
                            "read_wave_opt",
                            f"File {ss.read_wave_opt} does not exist",
                            None,
                        )
                    ],
                    self.Settings,
                )
        run_flags += setting_flag(ss.read_wave_opt)
        run_flags += setting_flag(ss.disp_tree)
        run_flags += setting_flag(ss.asserts)
        run_flags += setting_flag(ss.ieee_asserts)

        vpi = [] if ss.vpi is None else [ss.vpi] if not isinstance(ss.vpi, (list, tuple)) else list(ss.vpi)
        # TODO factor out cocotb handling
        if design.tb.cocotb and self.cocotb:
            vpi.append(self.cocotb.vpi_path())
            # tb_generics = list(design.tb.generics)  # TODO pass to cocotb?
            design.tb.generics = design.rtl.generics
        run_flags += setting_flag(vpi)

        if ss.debug:
            run_flags.extend(
                [
                    "--trace-processes",
                    "--checks",
                ]
            )

        run_flags += setting_flag(ss.stop_time)
        run_flags += setting_flag(ss.stop_delta)

        run_flags.extend(ss.generics_flags(design.tb.generics))

        x = self.elaborate(design.sim_sources, design.tb.top, design.language.vhdl)
        design.tb.top = x
        assert self.cocotb
        self.ghdl.run(
            "run",
            *cf,
            *design.sim_tops,
            *run_flags,
            env=self.cocotb.env(design),
        )

    def parse_reports(self) -> bool:
        success = True
        assert isinstance(self.settings, self.Settings)
        ss = self.settings

        dump_file = ss.wave or ss.vcd or ss.fst
        if dump_file:
            print("dump_file=", dump_file)
            opt_file = ss.read_wave_opt or ss.write_wave_opt
            extra_top = "top" if ss.wave else None
            signals, root_group = _get_wave_opt_signals(opt_file, extra_top)
            if dump_file == ss.fst:  # fst removes the common hierarchy
                signals = [s[len(root_group) :] for s in signals]
                root_group = []
            gen_gtkw(dump_file, signals, root_group)

        # TODO move
        if self.cocotb and self.design.tb and self.design.tb.cocotb:
            success &= self.cocotb.add_results(self.results)
        return success



import hashlib
import logging
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Tuple, Type, TypeVar, Union

from .dataclass import (
    Extra,
    Field,
    ValidationError,
    XedaBaseModel,
    root_validator,
    validation_errors,
    validator,
)
from .utils import WorkingDirectory, toml_load

log = logging.getLogger(__name__)

__all__ = [
    "Design",
    "DesignSource",
    "FileResource",
    "VhdlSettings",
    "LanguageSettings",
    "Clock",
]


class DesignValidationError(Exception):
    def __init__(
        self,
        errors: List[Tuple[str, str, str]],
        data: Dict[str, Any],
        *args: object,
        design_root: Union[None, str, os.PathLike] = None,
        design_name: Optional[str] = None,
        file: Optional[str] = None,
    ) -> None:
        super().__init__(*args)
        self.errors = errors  # location, msg, type/context
        self.data = data
        self.design_root = design_root
        self.design_name = design_name
        self.file = file

    def __str__(self) -> str:
        name = self.design_name or self.data.get("name")
        return "{}: {} error{} validating design{}:\n{}".format(
            self.__class__.__qualname__,
            len(self.errors),
            "s" if len(self.errors) > 1 else "",
            f" {name}" if name else "",
            "\n".join(
                "{}{} ({})\n".format(f"{loc}:\n   " if loc else "", msg, ctx)
                for loc, msg, ctx in self.errors
            ),
        )


class FileResource:
    def __init__(
        self,
        path: Union[str, os.PathLike, Dict[str, str]],
        _root_path: Optional[Path] = None,
        **data: Any,
    ) -> None:
        self._content_hash: Optional[str] = None
        try:
            if isinstance(path, dict):
                if "file" not in path:
                    raise ValueError("Required field 'file' is missing.")
                path = path["file"]
            p = Path(path)
            if not p.is_absolute():
                if not _root_path:
                    _root_path = Path.cwd()
                p = _root_path / p
            self.file = p.resolve(strict=True)
        except FileNotFoundError as e:
            log.critical("Design resource '%s' does not exist!", path)
            raise e from None

    @property
    def hash(self) -> str:
        """return hash of file content"""
        if self._content_hash is None:
            with open(self.file, "rb") as f:
                self._content_hash = hashlib.sha256(f.read()).hexdigest()
        return self._content_hash

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, FileResource):
            return False
        return self.hash == other.hash and self.file.samefile(other.file)

    def __hash__(self) -> int:
        # path is already absolute
        return hash((self.hash, str(self.file)))

    def __str__(self) -> str:
        return str(self.file)

    def __repr__(self) -> str:
        return "FileResource:" + self.__str__()


class DesignSource(FileResource):
    def __init__(
        self,
        path: Union[str, os.PathLike, Dict[str, str]],
        typ: Optional[str] = None,
        standard: Optional[str] = None,
        variant: Optional[str] = None,
        _root_path: Optional[Path] = None,
        **data: Any,
    ) -> None:
        super().__init__(path, _root_path, **data)

        def type_from_suffix(path: Path) -> Tuple[Optional[str], Optional[str]]:
            type_variants_map = {
                ("vhdl", variant): ["vhd", "vhdl"],
                ("verilog", variant): ["v"],
                ("systemverilog", variant): ["sv"],
                ("bsv", variant): ["bsv"],
                ("bs", variant): ["bs"],
            }
            for h, suffixes in type_variants_map.items():
                if path.suffix[1:] in suffixes:
                    return h
            return None, None

        self.type, self.variant = (typ, variant) if typ else type_from_suffix(self.file)
        if standard and len(standard) == 4:
            if standard.startswith("20") or standard.startswith("19"):
                standard = standard[2:]
        self.standard = standard

    def __eq__(self, other: Any) -> bool:  # pylint: disable=useless-super-delegation
        # added attributes do not change semantic equality
        return super().__eq__(other)

    def __hash__(self) -> int:  # pylint: disable=useless-super-delegation
        # added attributes do not change semantic identity
        return super().__hash__()


DefineType = Any
# the order matters!
# Union[FileResource, int, bool, float, str]

# Tuple -> Tuple[()] (empty tuple), but pydantic 1.9.0 + Python 3.9 typing do not like it
# Tuple of 0, 1, or 2 strings:
Tuple012 = Union[Tuple[str, ...], Tuple[str], Tuple[str, str]]  # xtype: ignore


class DVSettings(XedaBaseModel):  # type: ignore
    """Design/Verification settings"""

    sources: List[DesignSource]
    generics: Dict[str, DefineType] = Field(
        default={},
        description="Top-level generics/defines specified as a mapping",
        alias="parameters",
        allow_population_by_field_name=True,
        has_alias=True,
    )  # top defines/generics
    parameters: Dict[str, DefineType] = Field(
        default={},
        description="Top-level generics/defines specified as a mapping",
        alias="generics",
        allow_population_by_field_name=True,
        has_alias=True,
    )
    defines: Dict[str, DefineType] = Field(default={})

    # pylint: disable=no-self-argument
    @root_validator()
    def the_root_validator(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        value = values.get("parameters")
        if not value:
            value = values.get("generics")
        if value:
            for k, v in value.items():
                if isinstance(v, dict) and "file" in v:
                    value[k] = FileResource(v).__str__()
            values["generics"] = value
            values["parameters"] = value
        return values

    @validator("sources", pre=True, always=False)
    def sources_to_files(
        cls, sources: List[Union[DesignSource, str, os.PathLike, Path, Dict[str, Any]]]
    ) -> List[DesignSource]:
        ds = []
        for src in sources:
            if not isinstance(src, DesignSource):
                try:
                    src = DesignSource(src)
                except FileNotFoundError as e:
                    raise ValueError(
                        f"Source file: {src} was not found: {e.strerror} {e.filename}"
                    ) from None
            ds.append(src)
        return ds


class Clock(XedaBaseModel):
    port: Optional[str]


class RtlSettings(DVSettings):
    top: str = Field(
        description="Toplevel RTL module/entity",
    )
    clock: Clock = Clock(port=None)  # TODO rename to primary_clock?
    clocks: Dict[str, Clock] = {}
    clock_port: Optional[str] = None  # TODO remove?

    @root_validator(pre=False)
    def rtl_settings_validate(cls, values):
        clock = values.get("clock")
        clock_port = values.get("clock_port")
        if not clock.port:
            clock.port = clock_port
        if not clock_port:
            values["clock_port"] = clock.port
        if not values.get("clocks"):
            values["clocks"] = {"main_clock": clock}
        return values


class TbSettings(DVSettings):
    top: Tuple012 = Field(
        tuple(),
        description="Toplevel testbench module(s), specified as a tuple of strings. In addition to the primary toplevel, a secondary toplevel module can also be specified.",
    )
    sources: List[DesignSource] = []
    uut: Optional[str] = Field(
        None, description="instance name of the unit under test in the testbench"
    )
    cocotb: bool = Field(False, description="testbench is based on cocotb framework")

    @validator("top", pre=True)
    def top_validator(cls, top: Union[None, str, Sequence[str], Tuple012]) -> Tuple012:
        if top:
            if isinstance(top, str):
                return (top,)
            if isinstance(top, (tuple, list, Sequence)):
                assert len(top) <= 2
                return tuple(top)
        return tuple()


class LanguageSettings(XedaBaseModel):
    standard: Optional[str] = Field(
        None,
        description="Standard version",
        alias="version",
        has_alias=True,
        allow_population_by_field_name=True,
    )
    version: Optional[str] = Field(
        None,
        description="Standard version",
        alias="standard",
        has_alias=True,
        allow_population_by_field_name=True,
    )

    @validator("standard", pre=True)
    def two_digit_standard(cls, standard, values):
        if isinstance(standard, int):
            standard = str(standard)
        elif not isinstance(standard, str):
            raise ValueError("standard should be of type string")
        if standard and len(standard) == 4:
            if standard.startswith("20") or standard.startswith("19"):
                standard = standard[2:]
        return standard

    @root_validator(pre=True)
    def language_root_validator(cls, values):
        if "standard" in values:
            values["version"] = values["standard"]
        return values


class VhdlSettings(LanguageSettings):
    synopsys: bool = False


class Language(XedaBaseModel):
    vhdl: VhdlSettings = VhdlSettings()  # type: ignore
    verilog: LanguageSettings = LanguageSettings()  # type: ignore


class Design(XedaBaseModel):
    name: str
    rtl: RtlSettings
    tb: TbSettings = TbSettings()  # type: ignore
    language: Language = Language()

    class Config(XedaBaseModel.Config):
        extra = Extra.allow

    def __init__(
        self,
        design_root: Union[None, str, os.PathLike] = None,
        **data: Any,
    ) -> None:
        with WorkingDirectory(design_root):
            try:
                super().__init__(**data)
            except ValidationError as e:
                raise DesignValidationError(
                    validation_errors(e.errors()), data=data, design_root=design_root  # type: ignore
                ) from None

    @property
    def sim_sources(self) -> List[DesignSource]:
        if not self.tb:
            return []
        return self.rtl.sources + [
            src
            for src in self.tb.sources
            if src not in self.rtl.sources and src.type in ("vhdl", "verilog")
        ]

    @property
    def sim_tops(self) -> Tuple012:
        if self.tb:
            if self.tb.cocotb and self.rtl.top:
                return (self.rtl.top,)
            if self.tb.top is not None:
                return self.tb.top
        return tuple()

    T = TypeVar("T", bound="Design")

    @classmethod
    def from_toml(
        cls: Type[T],
        design_file: Union[str, os.PathLike],
        design_root: Union[None, str, os.PathLike] = None,
    ) -> T:
        """Load and validate a design description from TOML file"""
        if not isinstance(design_file, Path):
            design_file = Path(design_file)
        design_dict = toml_load(design_file)
        # Default value for design_root is the folder containing the design description file.
        if design_root is None:
            design_root = design_file.parent
        try:
            return cls(design_root=design_root, **design_dict)
        except DesignValidationError as e:
            raise DesignValidationError(  # add design_file to the emitted exception
                e.errors,
                data=e.data,
                design_root=e.design_root,
                design_name=e.design_name,
                file=str(design_file),
            ) from None



#!/usr/bin/env python3

import logging
import os
from pathlib import Path

from xeda import Design, Flow
from xeda.flow_runner import DefaultRunner
from xeda.flows import GhdlSim, Yosys
from xeda.tool import ExecutableNotFound

log = logging.getLogger(__name__)

SCRIPT_DIR = Path(__file__).parent.resolve()

design = Design.from_toml(SCRIPT_DIR / "sqrt.toml")
xeda_runner = DefaultRunner()


def test_sqrt() -> None:
    LONG_TEST = [8, 9, 16, 17, 18, 20, 21, 31, 32, 63, 64, 66]
    QUICK_TEST = [8, 32]
    ws = QUICK_TEST
    if os.environ.get("LONG_TEST"):
        print("runing long test!")
        ws = LONG_TEST
    for w in ws:
        assert design.tb
        design.tb.parameters = {**design.tb.parameters, "G_IN_WIDTH": w}
        os.environ["NUM_TV"] = str(100)
        f: Flow = xeda_runner.run_flow(GhdlSim, design)
        if not f.succeeded:
            raise AssertionError(f"test failed for w={w}")


def test_sqrt_yosys_synth() -> None:
    design.tb.parameters["G_IN_WIDTH"] = 32
    try:
        f = xeda_runner.run_flow(
            Yosys, design, {"fpga": {"part": "LFE5U-25F-6BG381C"}, "clock_period": 10.0}
        )
        assert f.succeeded
    except ExecutableNotFound:
        log.critical("Yosys executable not found")


if __name__ == "__main__":
    test_sqrt()
    test_sqrt_yosys_synth()



import logging
import os
from typing import Optional, List, Union
from urllib.request import urlretrieve
from urllib.parse import urlparse
from urllib.error import HTTPError

from .yosys import Yosys
from .flow import FlowFatalError, FpgaSynthFlow
from ..tool import Tool
from ..dataclass import Field
from ..board import WithFpgaBoardSettings, get_board_file_path, get_board_data

__all__ = ["Nextpnr"]

log = logging.getLogger(__name__)


class Nextpnr(FpgaSynthFlow):
    class Settings(WithFpgaBoardSettings):
        lpf_cfg: Optional[str] = None
        clock_period: float
        seed: Optional[int] = None
        random_seed: bool = False
        timing_allow_fail: bool = False
        ignore_loops: bool = Field(
            False, description="ignore combinational loops in timing analysis"
        )

        textcfg: Optional[str] = "config.txt"
        out_of_context: bool = Field(
            False,
            description="disable IO buffer insertion and global promotion/routing, for building pre-routed blocks",
        )
        lpf_allow_unconstrained = Field(
            False,
            description="don't require LPF file(s) to constrain all IOs",
        )
        extra_args: Optional[List[str]] = None
        py_script: Union[None, str, os.PathLike] = None
        out_json: Optional[str] = None
        sdf: Optional[str] = None
        log_to_file: Optional[str] = None
        report: Optional[str] = "report.json"
        detailed_timing_report: bool = True
        placed_svg: Optional[str] = "placed.svg"
        routed_svg: Optional[str] = "routed.svg"

    def init(self) -> None:
        assert isinstance(self.settings, self.Settings)
        ss = self.settings
        self.add_dependency(
            Yosys,
            Yosys.Settings(
                fpga=ss.fpga,
                clock_period=ss.clock_period,
            ),  # pyright: reportGeneralTypeIssues=none
        )

    def run(self) -> None:
        assert isinstance(self.settings, self.Settings)
        ss = self.settings
        yosys_flow = self.completed_dependencies[0]
        assert isinstance(yosys_flow, Yosys)
        assert isinstance(yosys_flow.settings, Yosys.Settings)
        assert yosys_flow.settings.netlist_json
        netlist_json = yosys_flow.run_path / yosys_flow.settings.netlist_json
        fpga_family = ss.fpga.family if ss.fpga.family else "generic"
        assert fpga_family in {
            "generic",
            "ecp5",
            "ice40",
            "nexus",
            "gowin",
            "fpga-interchange",
            "xilinx",
        }, "unsupported fpga family"
        next_pnr = Tool(f"nextpnr-{fpga_family}")

        assert (
            netlist_json.exists()
        ), f"netlist json file {netlist_json} does not exist!"

        lpf_cfg_path = ss.lpf_cfg
        board_data = get_board_data(ss.board)
        if not lpf_cfg_path and board_data and "lpf" in board_data:
            lpf = board_data["lpf"]
            r = urlparse(lpf)
            if r.scheme and r.netloc:
                try:
                    lpf_cfg_path, _ = urlretrieve(lpf)
                except HTTPError as e:
                    log.critical(
                        "Unable to retrive file from %s (HTTP Error %d)",
                        lpf,
                        e.code,
                    )
                    raise FlowFatalError("Unable to retreive LPF file") from None
            else:
                if "name" in board_data:
                    board_name = board_data["name"]
                else:
                    board_name = ss.board
                lpf_cfg_path = get_board_file_path(f"{board_name}.lpf")
        assert lpf_cfg_path

        freq_mhz = 1000 / ss.clock_period

        args = [
            "-q",
            "--json",
            netlist_json,
            "--freq",
            freq_mhz,
        ]
        if self.design.rtl.top:
            args.extend(["--top", self.design.rtl.top])
        if ss.seed is not None:
            args.extend(["--seed", ss.seed])

        package = ss.fpga.package
        speed = ss.fpga.speed
        device_type = ss.fpga.type
        if speed:
            args.extend(["--speed", speed])
        if ss.fpga.vendor and ss.fpga.vendor.lower() == "lattice":
            if package:
                package = package.upper()
                if package == "BG":
                    package = "CABGA"
                elif package == "MG":
                    package = "CSFBGA"
                assert ss.fpga.pins
                package += str(ss.fpga.pins)
        if ss.fpga.capacity:
            if device_type and device_type.lower() != "u":
                args.append(f"--{device_type}-{ss.fpga.capacity}")
            else:
                args.append(f"--{ss.fpga.capacity}")
        if ss.out_of_context:
            args.append("--out-of-context")
        if ss.lpf_allow_unconstrained:
            args.append("--lpf-allow-unconstrained")
        if ss.debug:
            args.append("--debug")
        if ss.verbose:
            args.append("--verbose")
        if ss.quiet:
            args.append("--quiet")
        if ss.random_seed:
            args.append("--randomize-seed")
        if ss.timing_allow_fail:
            args.append("--timing-allow-fail")
        if ss.ignore_loops:
            args.append("--ignore-loops")
        if ss.py_script:
            args += ["--run", ss.py_script]
        if package:
            args += ["--package", package]
        if lpf_cfg_path:
            # FIXME check what to do if no board
            args += ["--lpf", lpf_cfg_path]
        if ss.textcfg:
            args += ["--textcfg", ss.textcfg]
        if ss.out_json:
            args += ["--write", ss.out_json]
        if ss.nthreads:
            args += ["--threads", ss.nthreads]
        if ss.sdf:
            args += ["--sdf", ss.sdf]
        if ss.log_to_file:
            args += ["--log", ss.log_to_file]
        if ss.report:
            args += ["--report", ss.report]
        if ss.placed_svg:
            args += ["--placed-svg", ss.placed_svg]
        if ss.routed_svg:
            args += ["--routed-svg", ss.routed_svg]
        if ss.detailed_timing_report:
            args.append("--detailed-timing-report")
        if ss.extra_args:
            args += ss.extra_args
        next_pnr.run(*args)



from . import cli, design, flows, flow_runner
from .version import __version__
from .design import Design
from .flows.flow import Flow, SimFlow, SynthFlow, FPGA

__all__ = [
    "__version__",
    "cli",
    "design",
    "flows",
    "flow_runner",
    "Design",
    "Flow",
    "SimFlow",
    "SynthFlow",
    "FPGA",
]



import logging
import os
from functools import cached_property
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional

from junitparser import JUnitXml

from ..dataclass import Field, XedaBaseModel, validator
from ..design import Design
from ..tool import Tool

log = logging.getLogger(__name__)


class CocotbSettings(XedaBaseModel):
    coverage: bool = Field(
        False, description="Collect coverage data if supported by simulation tool."
    )
    reduced_log_fmt: bool = Field(
        True, description="Display shorter log lines in the terminal."
    )
    results_xml: str = Field(
        "results.xml",
        description="xUnit-compatible cocotb result file.",
        hidden_from_schema=True,
    )
    resolve_x: Literal["VALUE_ERROR", "ZEROS", "ONES", "RANDOM"] = Field(
        "VALUE_ERROR",
        description="how to resolve bits with a value of X, Z, U or W when being converted to integer.",
    )
    testcase: List[str] = Field(
        [],
        description="List of test-cases to run. Can also be specified as a comma-separated string. Currently used for cocotb testbenches only.",
    )
    random_seed: Optional[int] = Field(
        None,
        description="Seed the Python random module to recreate a previous test stimulus.",
    )
    gpi_extra: List[str] = Field(
        [],
        description="A comma-separated list of extra libraries that are dynamically loaded at runtime.",
    )

    @validator("testcase", "gpi_extra", pre=True, always=True)
    def str_to_list(cls, value):
        if isinstance(value, str):
            value = [s.strip() for s in value.split(",")]
        return value


class Cocotb(CocotbSettings, Tool):
    """Cocotb support for a SimFlow"""

    executable = "cocotb-config"
    sim_name: str

    """Not a stand-alone tool, but is used from a SimFlow"""

    def vpi_path(self) -> str:
        so_ext = "so"  # TODO windows?
        if self.version_gte(1, 6):
            so_path = self.run_get_stdout(
                "--lib-name-path",
                "vpi",
                self.sim_name,
            )
        else:
            so_path = (
                self.run_get_stdout("--prefix")
                + f"/cocotb/libs/libcocotbvpi_{self.sim_name}.{so_ext}"
            )

        log.info("cocotb.vpi_path: %s", so_path)
        return so_path

    def env(self, design: Design) -> Dict[str, Any]:
        ret = {}
        if design.tb and design.tb.cocotb:
            assert design.tb.top, "tb.top was not set by the parent SimFlow"
            coco_module = design.tb.sources[0].file.stem
            tb_top_path = design.tb.sources[0].file.parent
            ppath = []
            current_ppath = os.environ.get("PYTHONPATH")
            if current_ppath:
                ppath = current_ppath.split(os.pathsep)
            ppath.append(str(tb_top_path))
            top: str = (
                design.tb.top if isinstance(design.tb.top, str) else design.tb.top[0]
            )
            ret = {
                "MODULE": coco_module,
                "TOPLEVEL": top,  # TODO
                "COCOTB_REDUCED_LOG_FMT": int(self.reduced_log_fmt),
                "PYTHONPATH": os.pathsep.join(ppath),
                "COCOTB_RESULTS_FILE": self.results_xml,
                "COCOTB_RESOLVE_X": self.resolve_x,
            }
            if self.coverage:
                ret["COVERAGE"] = 1
            if self.testcase:
                ret["TESTCASE"] = ",".join(self.testcase)
            if self.random_seed is not None:
                ret["RANDOM_SEED"] = self.random_seed
            if self.gpi_extra:
                ret["GPI_EXTRA"] = ",".join(self.gpi_extra)
            log.info("Cocotb env: %s", ret)
        return ret

    @cached_property
    def _results(self):
        results_xml = self.results_xml
        if not Path(results_xml).exists():
            return JUnitXml()
        return JUnitXml.fromfile(results_xml)

    @property
    def results(self):
        return self._results

    @property
    def result_testcases(self):
        for ts in self.results:
            if ts is not None:
                for tc in ts:
                    if tc is not None:
                        yield {
                            "name": tc.name,
                            "result": str(tc),
                            "classname": tc.classname,
                            "time": round(tc.time, 3),
                        }

    def add_results(
        self, flow_results: Dict[str, Any], prefix: str = "cocotb."
    ) -> bool:
        """adds cocotb results to parent flow's results. returns success status"""
        xml = self.results
        flow_results[prefix + "tests"] = xml.tests
        flow_results[prefix + "errors"] = xml.errors
        flow_results[prefix + "failures"] = xml.failures
        flow_results[prefix + "skipped"] = xml.skipped
        flow_results[prefix + "time"] = xml.time
        failed = False
        if not xml.tests:
            failed = True
            log.error("No tests were discovered")
        if xml.errors:
            failed = True
            log.error("Cocotb: %d error(s)", xml.errors)
        if xml.failures:
            failed = True
            log.critical("Cocotb: %d failure(s)", xml.failures)
        if failed:
            flow_results["success"] = False

        return not failed



from pathlib import Path
from xeda.flows.ghdl import gen_gtkw, _get_wave_opt_signals, common_root


TESTS_DIR = Path(__file__).parent.absolute()
RESOURCES_DIR = TESTS_DIR / "resources"

def test_gen_gtkw():
    opt_file = RESOURCES_DIR / "wave.opt"
    dump_file = "debug.ghw"
    extra_top = "top"

    signals, root_group = _get_wave_opt_signals(opt_file, extra_top)
    assert len(signals) > 1
    assert len(root_group) == 2
    gen_gtkw(dump_file, signals, root_group)
    gtkw = Path("debug.gtkw")
    assert gtkw.exists()
    with open(gtkw) as f:
        lines = f.read().splitlines()
    for s in signals:
        sig_name = ".".join(s)
        assert lines.count(sig_name) > 0 , f"{sig_name} not in lines"


def test_common_root():
    cr = common_root([[1, 2, 3, 4, 5], [1, 2, 3, 7], [1, 2, 3], [1, 2, 3, 9]])
    assert cr == [1, 2, 3]


if __name__ == "__main__":
    test_gen_gtkw()



import os
import tempfile
from pathlib import Path

from xeda import Design
from xeda.flow_runner import DefaultRunner
from xeda.flows import IseSynth
from xeda.flows.flow import FPGA

TESTS_DIR = Path(__file__).parent.absolute()
RESOURCES_DIR = TESTS_DIR / "resources"
EXAMPLES_DIR = TESTS_DIR.parent / "examples"

os.environ["PATH"] += os.pathsep + os.path.join(TESTS_DIR, "fake_tools")


def test_ise_synth_py() -> None:
    path = RESOURCES_DIR / "design0/design0.toml"
    # Append to PATH so if the actual tool exists, would take precedences.
    assert path.exists()
    design = Design.from_toml(EXAMPLES_DIR / "vhdl" / "sqrt" / "sqrt.toml")
    settings = dict(fpga=FPGA("xc7a12tcsg325-1"), clock_period=5.5)
    with tempfile.TemporaryDirectory() as run_dir:
        print("Xeda run dir: ", run_dir)
        xeda_runner = DefaultRunner(run_dir, debug=True)
        flow = xeda_runner.run_flow(IseSynth, design, settings)
        settings_json = flow.run_path / "settings.json"
        results_json = flow.run_path / "results.json"
        assert settings_json.exists()
        assert results_json.exists()
        # assert flow.succeeded


if __name__ == "__main__":
    test_ise_synth_py()



#!/usr/bin/env python3

import inspect
import logging
import os
from pathlib import Path
from time import sleep
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Protocol,
    Union,
    runtime_checkable,
)
from zipfile import ZipFile

import click
from xeda.utils import XedaBaseModel

log = logging.getLogger()

RESOURCE_DIR = Path(__file__).parent.absolute() / "resource"


def write_file(path, data):
    path.parent.mkdir(parents=True, exist_ok=True)
    if isinstance(data, bytes):
        with open(path, "wb") as f:
            f.write(data)
    else:
        if data is None:
            data = []
        with open(path, "w") as f:
            if isinstance(data, list):
                f.writelines(data)
            else:
                f.write(data)


@runtime_checkable
class Executer(Protocol):
    def __call__(self, **kwargs: Any) -> int:
        ...


class WriteFile(Executer):
    def __init__(
        self,
        path: Union[str, os.PathLike],
        data: Union[None, List[str], str, bytes] = None,
        **kwargs: Any,
    ) -> None:
        if not isinstance(path, Path):
            path = Path(path)
        self.path = path
        self.data = data
        super().__init__(**kwargs)

    def __call__(self, **kwargs) -> int:
        write_file(self.path, self.data)
        return 0


class TouchFiles(Executer):
    def __init__(self, *paths: Union[str, os.PathLike], **kwargs) -> None:
        self.paths = paths
        super().__init__(**kwargs)

    def __call__(self, **kwargs) -> int:
        for path in self.paths:
            path = Path(path)
            path.parent.mkdir(parents=True, exist_ok=True)
            path.touch(exist_ok=True)
        return 0


class FakeTool(XedaBaseModel):
    version: Optional[str] = None
    version_template: Optional[str] = None
    vendor: Optional[str] = None
    help_options: list = ["--help"]
    version_options: list = ["--version"]
    options: dict = {}  # param_decls -> attrs
    arguments: dict = {}  # Dict[str, Optional[Dict[str, Any]]] = {}
    execute_: Executer = lambda **_kwargs: 0

    @property
    def version_banner(self) -> str:
        if self.version_template:
            return inspect.cleandoc(self.version_template.format(**(self.dict())))
        return "unknown"

    def execute(self, **kwargs) -> int:
        return self.execute_(**kwargs)


class FakeVivado(FakeTool):
    vendor = "Xilinx, Inc."
    version = "v2021.2"
    version_template = """Vivado {version} (64-bit)
        SW Build 3367213 on Tue Oct 19 02:47:39 MDT 2021
        IP Build 3369179 on Thu Oct 21 08:25:16 MDT 2021
        Copyright 1986-2021 {vendor} All Rights Reserved.
    """
    help_options = ["-help"]
    version_options = ["-version"]
    options = {
        "-mode": ["gui", "tcl", "batch"],
        "-init": dict(type=click.Path(exists=True)),
        "-source": dict(type=click.Path(exists=True)),
        "-verbose": None,
        "-nojournal": None,
        "-notrace": None,
        "-nolog": None,
    }
    arguments = {"project": dict(required=False)}

    def execute(self, **kwargs):
        print("cwd =", Path.cwd())
        tcl = kwargs.get("source")
        if tcl:
            sleep(0.3)
            with ZipFile(RESOURCE_DIR / "fake_vivado_reports") as zf:
                for file in zf.namelist():
                    if os.path.isdir(file):
                        continue
                    with zf.open(file) as rf:
                        data = rf.read()
                        write_file(Path("reports") / "route_design" / file, data)


fake_tools: Dict[str, FakeTool] = dict(
    vivado=FakeVivado(),  # type: ignore
    quartus_sh=FakeTool(
        options={"-t": dict(type=click.Path(exists=True), required=True)},
        execute_=TouchFiles(
            "reports/Flow_Summary.csv",
            "reports/Fitter/Resource_Section/Fitter_Resource_Utilization_by_Entity.csv",
            "reports/Timing_Analyzer/Multicorner_Timing_Analysis_Summary.csv",
        ),
    ),
    xtclsh=FakeTool(
        arguments={"script": dict(required=False, type=click.Path(exists=True))}
    ),
)

symlink_name = Path(__file__).stem

tool = fake_tools.get(symlink_name, FakeTool())


FC = Callable[..., Any]


def fake_tool_options(fake_tool: Optional[FakeTool]) -> FC:
    def decorator(f: FC) -> FC:
        if fake_tool:
            f = click.group(
                invoke_without_command=True,
                context_settings=dict(help_option_names=fake_tool.help_options),
            )(f)
            f = click.version_option(
                fake_tool.version,
                *fake_tool.version_options,
                message=fake_tool.version_banner,
            )(f)
            for arg, attrs in fake_tool.arguments.items():
                if attrs is None:
                    attrs = {}
                f = click.argument(arg, **attrs)(f)
            for param_decls, param_attrs in fake_tool.options.items():
                if isinstance(param_decls, str):
                    param_decls = (param_decls,)
                if param_attrs is None:
                    param_attrs = dict(is_flag=True)
                elif isinstance(param_attrs, list):
                    param_attrs = dict(type=click.Choice(param_attrs))
                elif isinstance(param_attrs, type):
                    param_attrs = dict(type=param_attrs)
                f = click.option(*param_decls, **param_attrs)(f)
        return f

    return decorator


@fake_tool_options(tool)
@click.pass_context
def cli(ctx: click.Context, **kwargs):
    if tool:
        print(f"Fake {ctx.info_name} kwargs:{kwargs} args:{ctx.args}")
        tool.execute(**kwargs)


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



import importlib
import json
import logging
import os
import re
from contextlib import AbstractContextManager
from copy import deepcopy
from datetime import datetime
from functools import cached_property, reduce
from pathlib import Path
from types import TracebackType
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    OrderedDict,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from .dataclass import XedaBaseModel

try:
    import tomllib  # type: ignore # pyright: reportMissingImports=none
except ModuleNotFoundError:
    # python_version < "3.11":
    import tomli as tomllib  # type: ignore

__all__ = [
    "SDF",
    # utility functions
    "toml_load",
    "toml_loads",
    "cached_property",
    "unique",
    "WorkingDirectory",
]

log = logging.getLogger(__name__)


T = TypeVar("T")


class WorkingDirectory(AbstractContextManager):
    def __init__(self, wd: Union[None, str, os.PathLike]):
        self.prev_wd: Optional[Path] = None
        if wd is not None and not isinstance(wd, Path):
            wd = Path(wd)
        self.wd = wd

    def __enter__(self) -> None:
        self.prev_wd = Path.cwd()
        if self.wd:
            log.debug("Changing working directory from %s to %s", self.prev_wd, self.wd)
            os.chdir(self.wd)

    def __exit__(
        self,
        exception_type: Optional[Type[BaseException]],
        exception_value: Optional[BaseException],
        exception_traceback: Optional[TracebackType],
    ) -> None:
        if self.prev_wd:
            log.debug("Changing back working directory to %s", self.prev_wd)
            os.chdir(self.prev_wd)


class SDF(XedaBaseModel):
    """
    root: region where the SDF is applied
        /TESTBENCH/UUT
    if tb.top=TESTBENCH and the simulation netlist is instantiated inside the testbench with an instance name of UUT
    """

    root: Optional[str] = None
    min: Optional[str] = None
    max: Optional[str] = None
    typ: Optional[str] = None

    def __init__(self, *args: str, **data: str) -> None:
        if args:
            assert len(args) == 1 and isinstance(args[0], str), "only 1 str argument"
            data["max"] = args[0]
        super().__init__(**data)

    def delay_items(self) -> Iterable[Tuple[str, Union[str, None]]]:
        """returns an iterable of (delay_type, sdf_file)"""
        return tuple(
            (delay_type, self.dict().get(delay_type))
            for delay_type in ("min", "max", "typ")
            if self.dict().get(delay_type)
        )


def toml_load(path: Union[str, os.PathLike]) -> Dict[str, Any]:  # type: ignore
    with open(path, "rb") as f:
        return tomllib.load(f)  # type: ignore


def toml_loads(s: str) -> Dict[str, Any]:
    return tomllib.loads(s)  # type: ignore


def backup_existing(path: Path) -> Optional[Path]:
    if not path.exists():
        log.warning("%s does not exist for backup!", path)
        return None
    modifiedTime = os.path.getmtime(path)
    suffix = (
        f'.backup_{datetime.fromtimestamp(modifiedTime).strftime("%Y-%m-%d-%H%M%S")}'
    )
    if path.suffix:
        suffix += path.suffix
    backup_path = path.with_suffix(suffix)
    typ = "file" if path.is_file() else "directory" if path.is_dir() else "???"
    log.warning(
        "Renaming existing %s from '%s' to '%s'", typ, path.name, backup_path.name
    )
    # TODO use shutil.move instead? os.rename vs Path.rename?
    # os.rename(path, backup_path)
    return path.rename(backup_path)


def dump_json(data: object, path: Path, backup_previous: bool = True) -> None:
    if path.exists() and backup_previous:
        backup_existing(path)
        assert not path.exists(), "Old file still exists!"

    with open(path, "w") as outfile:
        json.dump(
            data,
            outfile,
            default=lambda x: x.__dict__ if hasattr(x, "__dict__") else str(x),
            indent=4,
        )


def unique(lst: List[Any]) -> List[Any]:
    """uniquify list while preserving order"""
    return list(OrderedDict.fromkeys(lst))


def camelcase_to_snakecase(name: str) -> str:
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def snakecase_to_camelcase(name: str) -> str:
    return "".join(word.title() for word in name.split("_"))


def load_class(
    full_class_string: str, defualt_module_name: Optional[str] = None
) -> Optional[type]:
    cls_path_lst = full_class_string.split(".")
    assert len(cls_path_lst) > 0

    cls_name = snakecase_to_camelcase(cls_path_lst[-1])
    if len(cls_path_lst) == 1:  # module name not specified, use default
        mod_name = defualt_module_name
    else:
        mod_name = ".".join(cls_path_lst[:-1])
    assert mod_name

    module = importlib.import_module(
        mod_name, __package__ if mod_name.startswith(".") else None
    )
    cls = getattr(module, cls_name)
    if not isinstance(cls, type):
        return None
    return cls


def dict_merge(
    base_dict: Dict[Any, Any], merge_dict: Dict[Any, Any], add_new_keys: bool = True
) -> Dict[Any, Any]:
    """
    returns content of base_dict merge with content of merge_dict.
    if add_new_keys=False keys in merge_dict not existing in base_dict are ignored
    """
    rtn_dct = deepcopy(base_dict)
    if add_new_keys is False:
        merge_dict = {
            key: merge_dict[key] for key in set(rtn_dct).intersection(set(merge_dict))
        }

    rtn_dct.update(
        {
            key: dict_merge(rtn_dct[key], merge_dict[key], add_new_keys=add_new_keys)
            if isinstance(rtn_dct.get(key), dict) and isinstance(merge_dict[key], dict)
            else merge_dict[key]
            for key in merge_dict
        }
    )
    return rtn_dct


def try_convert(
    s: Any, convert_lists: bool = False
) -> Union[bool, int, float, str, List[Union[bool, int, float, str, List[Any]]]]:
    if s is None:
        return "None"
    if isinstance(s, str):
        s = s.strip()
        assert isinstance(s, str)
        if s.startswith('"') or s.startswith("'"):
            return s.strip("\"'")
        if convert_lists and s.startswith("[") and s.endswith("]"):
            s = re.sub(r"\s+", "", s)
            return try_convert(list(s.strip("][").split(",")))
        # Should NOT convert dict, set, etc!
        if re.match(r"^\d+$", s):
            return int(s)
        if s.lower() in ["true", "yes"]:
            return True
        if s.lower() in ["false", "no"]:
            return False
        try:
            return float(s)
        except ValueError:
            pass
        return s
    if isinstance(s, (int, float, bool)):
        return s
    if isinstance(s, (tuple)):
        s = list(s)
    if isinstance(s, (list)):
        return [try_convert(e) for e in s]
    return str(s)


def get_hierarchy(dct: Dict[str, Any], path, sep="."):
    if isinstance(path, str):
        path = path.split(sep)
    try:
        return reduce(dict.__getitem__, path, dct)
    except ValueError:
        return None


def set_hierarchy(dct: Dict[str, Any], path, value, sep="."):
    if isinstance(path, str):
        path = path.split(sep)
    k = path[0]
    if len(path) == 1:
        dct[k] = value
    else:
        if k not in dct:
            dct[k] = {}
        set_hierarchy(dct[k], path[1:], value, sep)


def append_flag(flag_list: List[str], flag: str) -> List[str]:
    if flag not in flag_list:
        flag_list.append(flag)
    return flag_list


def common_root(signals: List[List[T]]) -> List[T]:
    longest: Optional[List[T]] = None
    for sig in signals:
        if not sig:
            continue
        if longest is None:
            longest = sig[:-1]
            continue
        new_len = min(len(longest), len(sig))
        longest = longest[:new_len]
        if not longest:
            break
        for i in range(new_len):
            if sig[i] != longest[i]:
                longest = longest[:i]
                break
    return longest if longest else []



import contextlib
import logging
import os
import re
import subprocess
import sys
from pathlib import Path
from sys import stderr
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from .dataclass import Field, XedaBaseModeAllowExtra, XedaBaseModel, validator
from .utils import cached_property, unique

log = logging.getLogger(__name__)


__all__ = [
    "ToolException",
    "NonZeroExitCode",
    "ExecutableNotFound",
    "Docker",
    "Tool",
    "run_process",
]


class ToolException(Exception):
    """Super-class of all tool exceptions"""


class NonZeroExitCode(ToolException):
    def __init__(self, command_args: Any, exit_code: int, *args: object) -> None:
        self.command_args = command_args
        self.exit_code = exit_code
        super().__init__(*args)


class ExecutableNotFound(ToolException):
    def __init__(
        self, executable: str, tool: str, path: str, *args: Any, **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.exec = executable
        self.tool = tool
        self.path = path


class RemoteSettings(XedaBaseModel):
    enabled: bool = False
    junest_path: str  # FIXME REMOVE
    junest_method: str = "ns"
    junest_mounts: Dict[str, str] = {}
    exec_path: Optional[str] = None
    hostname: str
    port: int = 22


OptionalPath = Union[None, str, os.PathLike]
OptionalBoolOrPath = Union[None, bool, str, os.PathLike]


class Docker(XedaBaseModel):
    enabled: bool = False
    command: List[str] = []
    platform: Optional[str]
    image: str = Field(description="Docker image name")
    tag: str = Field("latest", description="Docker image tag")
    registry: Optional[str] = Field(None, description="Docker image registry")
    mounts: Dict[str, str] = {}

    # TODO this is only for a Linux container
    @cached_property
    def _cpuinfo(self) -> List[List[str]]:
        ret = self._run_docker("cat", "/proc/cpuinfo", stdout=True)
        assert ret
        return [x.split("\n") for x in re.split(r"\n\s*\n", ret, re.MULTILINE)]

    @property
    def cpuinfo(self) -> List[List[str]]:
        return self._cpuinfo

    @property
    def nproc(self) -> Optional[int]:
        return len(self._cpuinfo)

    @property
    def name(self) -> str:
        return self.command[0] if self.command else "_"

    def run(
        self,
        *args: Any,
        env: Optional[Dict[str, Any]] = None,
        stdout: OptionalBoolOrPath = None,
        check: bool = True,
        root_dir: OptionalPath = None,
    ) -> Union[None, str]:
        """Run the tool from a docker container"""
        cpuinfo_file = Path(".cpuinfo").resolve()
        with open(cpuinfo_file, "w") as f:
            for proc in self.cpuinfo:
                for line in proc:
                    assert isinstance(line, str)
                    if line.startswith("Features"):
                        line += " sse sse2"
                    f.write(line + "\n")
                f.write("\n")

        self.mounts[str(cpuinfo_file)] = "/proc/cpuinfo"
        return self._run_docker(
            *self.command, *args, env=env, stdout=stdout, check=check, root_dir=root_dir
        )

    def _run_docker(
        self,
        *args: Any,
        env: Optional[Dict[str, Any]] = None,
        stdout: OptionalBoolOrPath = None,
        check: bool = True,
        root_dir: OptionalPath = None,
    ) -> Union[None, str]:
        cwd = Path.cwd()
        wd = root_dir if root_dir else cwd
        if not isinstance(wd, Path):
            wd = Path(wd)
        docker_args = [
            "--rm",
            f"--workdir={wd}",
        ]
        self.mounts[str(wd)] = str(wd)
        self.mounts[str(cwd)] = str(cwd)
        if not stdout and sys.stdout.isatty():
            docker_args += ["--tty", "--interactive"]
        if self.platform:
            docker_args += ["--platform", self.platform]
        for k, v in self.mounts.items():
            docker_args.append(f"--volume={k}:{v}")
        if env:
            env_file = wd / f".{self.name}_docker.env"
            with open(env_file, "w") as f:
                f.write("\n".join(f"{k}={v}" for k, v in env.items()))
            docker_args.extend(["--env-file", str(env_file)])

        return run_process(
            "docker",
            ["run", *docker_args, f"{self.image}:{self.tag}", *args],
            env=None,
            stdout=stdout,
            check=check,
            tool_name=self.name,
        )


def run_process(
    executable: str,
    args: Optional[Sequence[Any]] = None,
    env: Optional[Dict[str, Any]] = None,
    stdout: OptionalBoolOrPath = None,
    check: bool = True,
    cwd: OptionalPath = None,
    tool_name: str = "",
) -> Union[None, str]:
    if args is None:
        args = []
    args = [str(a) for a in args]
    if env is not None:
        env = {k: str(v) for k, v in env.items()}
    log.info("Running `%s`", " ".join([executable, *args]))
    if cwd:
        log.info("cwd=%s", cwd)
    if stdout and isinstance(stdout, (str, os.PathLike)):
        stdout = Path(stdout)
        log.info("redirecting stdout to %s", stdout)

        def cm_call():
            assert stdout
            return open(stdout, "w")

        cm = cm_call
    else:
        cm = contextlib.nullcontext
    with cm() as f:
        try:
            with subprocess.Popen(
                [executable, *args],
                cwd=cwd,
                shell=False,
                stdout=f if f else subprocess.PIPE if stdout else None,
                bufsize=1,
                universal_newlines=True,
                encoding="utf-8",
                errors="replace",
                env=env,
            ) as proc:
                log.info("Started %s[%d]", executable, proc.pid)
                if stdout:
                    if isinstance(stdout, bool):
                        out, err = proc.communicate(timeout=None)
                        if check and proc.returncode != 0:
                            raise NonZeroExitCode(proc.args, proc.returncode)
                        if err:
                            print(err, file=stderr)
                        return out.strip()
                    log.info("Standard output is logged to: %s", stdout)
                else:
                    proc.wait()
        except FileNotFoundError as e:
            path = env["PATH"] if env and "PATH" in env else os.environ.get("PATH", "")
            raise ExecutableNotFound(e.filename, tool_name, path, *e.args) from None
    if check and proc.returncode != 0:
        raise NonZeroExitCode(proc.args, proc.returncode)
    return None


def fake_cpu_info(file=".xeda_cpuinfo", ncores=4):
    with open(file, "w") as f:
        for i in range(ncores):
            cpuinfo: Dict[str, Any] = {
                "processor": i,
                "vendor_id": "GenuineIntel",
                "family": 6,
                "model": 60,
                "core id": i,
                "cpu cores": ncores,
                "fpu": "yes",
                "model name": "Intel(R) Core(TM) i7-4790K CPU @ 4.00GHz",
                "flags": "fpu vme tsc msr pae mce cx8 mmx fxsr sse sse2 avx2",
            }
            for k, v in cpuinfo.items():
                ws = "\\t" * (1 if len(k) >= 8 else 2)
                f.write(f"{k}{ws}: {v}")


def _run_processes(commands: List[List[str]], cwd: OptionalPath = None) -> None:
    """Run a list commands to completion. Throws if any of them did not execute and exit normally"""
    # if args:
    #     args = [str(a) for a in args]
    # if env:
    #     env = {str(k): str(v) for k, v in env.items()}
    processes = []
    for cmd in commands:
        log.info("Running `%s`", " ".join(cmd))
        with subprocess.Popen(
            cmd,
            cwd=cwd,
            shell=False,
            #   stdout=subprocess.PIPE if stdout else None,
            bufsize=1,
            universal_newlines=True,
            encoding="utf-8",
            errors="replace",
            #   env=env
        ) as proc:
            assert isinstance(proc.args, list)
            log.info("Started %s[%d]", str(proc.args[0]), proc.pid)
            processes.append(proc)

    for p in processes:
        p.wait()

    for p in processes:
        if p.returncode != 0:
            raise Exception(f"Process exited with return code {p.returncode}")


class Tool(XedaBaseModeAllowExtra):
    """abstraction for an EDA tool"""

    executable: str
    minimum_version: Union[None, Tuple[Union[int, str], ...]] = None
    default_args: Optional[List[str]] = None

    remote: Optional[RemoteSettings] = Field(None, hidden_from_schema=True)
    docker: Optional[Docker] = Field(None, hidden_from_schema=True)
    log_stdout: bool = Field(
        False, description="Log stdout to a file", hidden_from_schema=True
    )
    log_stderr: bool = Field(
        False, description="Log stderr to a file", hidden_from_schema=True
    )
    bin_path: str = Field(
        None, description="Path to the tool binary", hidden_from_schema=True
    )

    def __init__(self, executable: Optional[str] = None, **kwargs):
        if executable:
            assert "executable" not in kwargs, "executable specified twice"
            kwargs["executable"] = executable
        super().__init__(**kwargs)

    @validator("docker", pre=True, always=True)
    def validate_docker(cls, value, values):
        if value and not value.command:
            value.command = [values.get("executable")]
        return value

    @cached_property
    def _info(self) -> Dict[str, str]:
        log.warning("Tool.info is not implemented for %s", self.__class__.__name__)
        return {}

    @property  # pydantic can't handle cached_property as private
    def info(self) -> Dict[str, str]:
        inf = self._info
        inf["version"] = ".".join(self.version)
        return inf

    @cached_property
    def _version(self) -> Tuple[str, ...]:
        out = self.run_get_stdout(
            "--version",
        )
        assert isinstance(out, str)
        so = re.split(r"\s+", out)
        version_string = so[1] if len(so) > 1 else so[0] if len(so) > 0 else ""
        return tuple(version_string.split("."))

    @property
    def version(self) -> Tuple[str, ...]:
        return self._version

    @staticmethod
    def _version_is_gte(v1: Tuple[str, ...], v2: Tuple[Union[int, str], ...]) -> bool:
        """version v1 is greater than or equal to v2"""
        for tv_s, sv_s in zip(v2, v1):
            try:
                tv = int(tv_s)
            except ValueError:
                tv = -1
            try:
                sv = int(sv_s)
            except ValueError:
                sv = -1
            if sv < tv:
                return False
            if sv > tv:
                return True
        return True

    def version_gte(self, *args: Union[int, str]) -> bool:
        """Tool version is greater than or equal to version specified in args"""
        return self._version_is_gte(self.version, args)

    def _run_process(
        self,
        args: Optional[Sequence[Any]] = None,
        env: Optional[Dict[str, Any]] = None,
        stdout: OptionalBoolOrPath = None,
        check: bool = True,
        cwd: OptionalPath = None,
    ):
        return run_process(
            self.executable,
            args,
            env=env,
            stdout=stdout,
            check=check,
            cwd=cwd,
            tool_name=self.__class__.__name__,
        )

    def _run_system(
        self,
        *args: Any,
        stdout: OptionalBoolOrPath = None,
        env: Optional[Dict[str, Any]] = None,
        cwd: OptionalPath = None,
        check: bool = True,
    ) -> Union[None, str]:
        """Run the tool if locally installed on the system and available on the current user's PATH"""
        if env is not None:
            env = {**os.environ, **env}
        return self._run_process(args, env=env, stdout=stdout, check=check, cwd=cwd)

    def _run_remote(
        self,
        *args: Any,
        env: Optional[Dict[str, Any]] = None,
        stdout: OptionalBoolOrPath = None,
        check: bool = True,
        cwd: OptionalPath = None,
    ) -> None:
        # FIXME remote execution is broken
        #
        # if env is not None:
        # env = {**os.environ, **env}
        assert self.remote, "tool.remote settings not available!"

        remote_sshfs_mount_dir = "~/mnt"
        wd = str(cwd) if cwd else os.curdir
        self.remote.junest_mounts[remote_sshfs_mount_dir] = wd
        mount_opts = [f"--bind {k} {v}" for k, v in self.remote.junest_mounts.items()]
        junest_backend_opts = ["-b"] + mount_opts
        executable = self.executable
        if self.remote.exec_path:
            executable = os.path.join(self.remote.exec_path, executable)
        junest_cmd = (
            [self.remote.junest_path]
            + junest_backend_opts
            + ["--", executable]
            + list(*args)
        )

        remote_cmd = junest_cmd
        client_nc_port = 34567
        reverse_sftp_port = 10000

        sshfs_opts = ["directport=10000", "idmap=user", "exec", "compression=yes"]

        sshfs_cmd = [
            "sshfs",
            "-o",
            ",".join(sshfs_opts),
            f"localhost:{wd}",
            remote_sshfs_mount_dir,
        ]

        ssh_proc = [
            "ssh",
            self.remote.hostname,
            "-p",
            str(self.remote.port),
        ]

        sshfs_proc = (
            ssh_proc
            + [
                "-R",
                f"{reverse_sftp_port}:localhost:{client_nc_port}",
                "mkdir",
                "-p",
                remote_sshfs_mount_dir,
                "&&",
            ]
            + sshfs_cmd
            + ["&&"]
            + remote_cmd
        )

        ncat_proc = [
            "ncat",
            "-l",
            "-p",
            f"{client_nc_port}",
            "-e",
            "/usr/libexec/sftp-server",
        ]
        _run_processes([sshfs_proc, ncat_proc])

    def _run(
        self,
        *args: Any,
        env: Optional[Dict[str, Any]] = None,
        stdout: OptionalBoolOrPath = None,
        check: bool = True,
    ) -> Union[None, str]:
        if self.default_args:
            args = tuple(unique(self.default_args + list(args)))
        if self.remote and self.remote.enabled:
            self._run_remote(*args, env=env, stdout=stdout, check=check)
            return None
        if self.docker and self.docker.enabled:
            return self.docker.run(*args, env=env, stdout=stdout, check=check)
        return self._run_system(*args, env=env, stdout=stdout, check=check)

    def run(self, *args: Any, env: Optional[Dict[str, Any]] = None) -> None:
        self._run(*args, env=env, stdout=None)

    def run_get_stdout(self, *args: Any, env: Optional[Dict[str, Any]] = None) -> str:
        out = self._run(*args, env=env, stdout=True)
        assert isinstance(out, str)
        return out

    def run_stdout_to_file(
        self,
        *args: Any,
        stdout: OptionalBoolOrPath,
        env: Optional[Dict[str, Any]] = None,
    ) -> None:
        self._run(*args, env=env, stdout=stdout)



# all Flow classes imported here can be used from FlowRunners and will be reported on the command-line help
from .dc import Dc
from .diamond import DiamondSynth
from .ghdl import GhdlSim, GhdlSynth
from .ise import IseSynth
from .modelsim import Modelsim
from .nextpnr import Nextpnr
from .openfpgaloader import Openfpgaloader
from .quartus import Quartus
from .vivado.vivado_synth import VivadoSynth
from .vivado.vivado_sim import VivadoSim
from .vivado.vivado_postsynthsim import VivadoPostsynthSim
from .vivado.vivado_power import VivadoPower
from .vivado.vivado_alt_synth import VivadoAltSynth
from .yosys import Yosys

# alias for backwards compatibility
from .yosys import Yosys as YosysSynth  # pylint: disable=reimported

__all__ = [
    "Dc",
    "DiamondSynth",
    "GhdlSim",
    "GhdlSynth",
    "IseSynth",
    "Modelsim",
    "Nextpnr",
    "Openfpgaloader",
    "Quartus",
    "VivadoAltSynth",
    "VivadoPostsynthSim",
    "VivadoPower",
    "VivadoSim",
    "VivadoSynth",
    "Yosys",
    "YosysSynth",
]



from subprocess import check_call


def quartus_synth():
    check_call(
        [
            "xeda",
            "run",
            "quartus",
            "--design",
            "pipelined_adder.toml",
            "--settings",
            "dockerized=true",
            "fpga.part=10CL016YU256C6G",
            "clock_period=15",
        ]
    )


if __name__ == "__main__":
    quartus_synth()



# © 2022 [Kamyar Mohajerani](mailto:kamyar@ieee.org)
"""Xeda Command-line interface"""
import inspect
import json
import logging
import multiprocessing
import os
import re
import sys
from dataclasses import dataclass
from datetime import datetime
from functools import reduce
from pathlib import Path
from typing import Optional, Sequence, Tuple

import click
import coloredlogs
import yaml
from click.shell_completion import get_completion_class
from rich import box
from rich.style import Style
from rich.table import Table
from rich.text import Text

from .cli_utils import discover_flow_class  # pylint: disable=no-name-in-module
from .cli_utils import ClickMutex, OptionEatAll, XedaHelpGroup, settings_to_dict
from .console import console
from .design import Design, DesignValidationError
from .flow_runner import DefaultRunner
from .flows.flow import Flow, FlowFatalError, FlowSettingsError, registered_flows
from .tool import ExecutableNotFound, NonZeroExitCode
from .utils import toml_load

log = logging.getLogger(__name__)


def load_xeda(file: Path):
    ext = file.suffix.lower()

    if ext == ".toml":
        return toml_load(file)
    with open(file) as f:
        if ext == ".json":
            return json.load(f)
        if ext == ".yaml":
            return yaml.safe_load(f)
        sys.exit(
            f"File {file} has unknown extension {ext}. Currently supported formats are TOML (.toml) and JSON (.json)"
        )


def load_design_from_toml(design_file) -> Design:
    return Design.from_toml(design_file)


def get_available_flows():
    """alternative method using inspect"""
    mod = "xeda.flows"
    fc = inspect.getmembers(
        sys.modules[mod],
        lambda cls: inspect.isclass(cls)
        and issubclass(cls, Flow)
        and not inspect.isabstract(cls),
    )
    return {n: (mod, cls) for n, cls in fc}


CONTEXT_SETTINGS = dict(
    auto_envvar_prefix="XEDA",
    help_option_names=["--help", "-h"],
    max_content_width=console.width,
)


@dataclass
class XedaOptions:
    verbose: bool = False
    quiet: bool = False
    debug: bool = False


def log_to_file(logdir: Path):
    timestamp = datetime.now().strftime("%Y-%m-%d-%H%M%S%f")[:-3]
    logdir.mkdir(exist_ok=True, parents=True)
    logfile = logdir / f"xeda_{timestamp}.log"
    log.info("Logging to %s", logfile)
    fileHandler = logging.FileHandler(logfile)
    logFormatter = logging.Formatter(
        "%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s"
    )
    fileHandler.setFormatter(logFormatter)
    log.root.addHandler(fileHandler)


@click.group(cls=XedaHelpGroup, no_args_is_help=True, context_settings=CONTEXT_SETTINGS)
@click.option("--verbose", is_flag=True, help="Enables verbose mode.")
@click.option("--quiet", is_flag=True, help="Enable quiet mode.")
@click.option(
    "--debug",
    show_envvar=True,
    is_flag=True,
    # type=DebugLevel,  # click.Choice([str(l.value) for l in DebugLevel]),
    # help=f"""
    #         Set debug level. DEBUG_LEVEL values corresponds to:
    #         {', '.join([f"{l.value}: {l.name}" for l in DebugLevel])}
    #     """,
)
@click.version_option(message="Xeda v%(version)s")
@click.pass_context
def cli(ctx: click.Context, **kwargs):
    ctx.obj = XedaOptions(**kwargs)

    log.setLevel(
        logging.WARNING
        if ctx.obj.quiet
        else logging.DEBUG
        if ctx.obj.debug
        else logging.INFO
    )
    coloredlogs.install(
        None, fmt="%(asctime)s %(levelname)s %(message)s", logger=log.root
    )


@cli.command(
    context_settings=CONTEXT_SETTINGS,
    short_help="Run a flow.",
    help="Run the flow identified by FLOW_NAME. A snake_case styled FLOW_NAME (e.g. ghdl_sim) is converted to a CamelCase class name (e.g. GhdlSim).",
    no_args_is_help=False,
)
@click.argument(
    "flow", metavar="FLOW_NAME", type=click.Choice(sorted(list(registered_flows)))
)
@click.option(
    "--xeda-run-dir",
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=True,
        allow_dash=True,
        path_type=Path,
    ),
    envvar="XEDA_RUN_DIR",
    help="Parent folder for execution of xeda commands.",
    default="xeda_run",
    show_default=True,
    show_envvar=True,
)
@click.option(
    "--cached-dependencies",
    is_flag=True,
    help="Don't run dependency flows if a previous successfull run on the same design and flow settings exists.",
)
@click.option(
    "--run_in_existing_dir",
    is_flag=True,
    help="DO NOT USE!",
    hidden=True,
)
# @click.option(
#     "--force-run",
#     is_flag=True,
#     help="Force re-run of flow and all dependencies, even if they are already up-to-date",
# )
@click.option(
    "--xedaproject",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
        allow_dash=False,
        path_type=Path,
    ),
    cls=ClickMutex,
    mutually_exclusive_with=["design_file"],
    help="Path to Xeda project file.",
)
@click.option(
    "--design-name",
    cls=ClickMutex,
    mutually_exclusive_with=["design_file"],
    help="Specify design.name in case multiple designs are available in a xedaproject.",
)
@click.option(
    "--design-file",
    "--design",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
        allow_dash=False,
        path_type=Path,
    ),
    cls=ClickMutex,
    mutually_exclusive_with=["xedaproject"],
    help="Path to Xeda design file containing the description of a single design.",
)
@click.option(
    "--flow-settings",
    "--settings",
    metavar="KEY=VALUE...",
    type=tuple,
    cls=OptionEatAll,
    help="""Override setting values for the executed flow. Separate multiple KEY=VALUE overrides with commas. KEY can be a hierarchical name using dot notation.
    Example: --settings clock_period=2.345 impl.strategy=Debug
    """
    #  examples: # FIXME move to docs
    # - xeda vivado_sim --flow-settings stop_time=100us
    # - xeda vivado_synth --flow-settings impl.strategy=Debug --flow-settings clock_period=2.345
)  # pylint: disable=C0116:missing-function-docstring
@click.pass_context
def run(
    ctx: click.Context,
    flow: str,
    cached_dependencies: bool = False,
    run_in_existing_dir: bool = False,
    # force_run: bool = False,
    xeda_run_dir: Optional[Path] = None,
    xedaproject: Optional[str] = None,
    design_name: Optional[str] = None,
    design_file: Optional[str] = None,
    flow_settings: Optional[Tuple[str, ...]] = None,
):
    assert ctx
    options: XedaOptions = ctx.obj or XedaOptions()
    assert xeda_run_dir
    log_to_file(xeda_run_dir / "Logs")
    # TODO get default flow configs?
    flows_config = {}
    if design_file:
        try:
            design = Design.from_toml(design_file)
        except DesignValidationError as e:
            log.critical("%s", e)
            sys.exit(1)
    elif xedaproject:
        toml_path = Path(xedaproject)
        try:
            xeda_project = load_xeda(toml_path)
        except FileNotFoundError:
            try:
                xeda_project = load_xeda(toml_path.parent / (toml_path.stem + ".json"))
            except FileNotFoundError:
                sys.exit(
                    f"Cannot open project file: {toml_path}. Please run from the project directory with xedaproject.toml or specify the correct path using the --xedaproject flag"
                )
        flows_config = xeda_project.get("flows", {})
        designs = xeda_project["design"]
        if not isinstance(designs, Sequence):
            designs = [designs]
        log.info(
            "Available designs: %s",
            ", ".join(d.get("name", "<NO_NAME>") for d in designs),
        )
        design_dict = {}
        if len(designs) == 1:
            design_dict = designs[0]
        elif design_name:
            for x in designs:
                if x["name"] == design_name:
                    design_dict = x
            if not design_dict:
                log.critical(
                    'Design "%s" not found in the current project.', design_name
                )
                sys.exit(1)
        design = Design(design_root=toml_path.parent, **design_dict)
    else:
        sys.exit("No design or project specified!")
    flow_overrides = settings_to_dict(flow_settings)
    log.debug("flow_overrides: %s", flow_overrides)
    flow_overrides = {**flows_config.get(flow, {}), **flow_overrides}
    flow_class = discover_flow_class(flow)
    runner = DefaultRunner(
        xeda_run_dir,
        debug=options.debug,
        cached_dependencies=cached_dependencies,
        run_in_existing_dir=run_in_existing_dir,
    )
    try:
        runner.run_flow(flow_class, design, flow_overrides)
    except FlowFatalError as e:
        log.critical(
            "Flow %s failed: FlowFatalException %s",
            flow,
            " ".join(str(a) for a in e.args),
        )
        sys.exit(1)
    except NonZeroExitCode as e:
        log.critical(
            "Flow %s failed: NonZeroExitCode %s", flow, " ".join(str(a) for a in e.args)
        )
        sys.exit(1)
    except ExecutableNotFound as e:
        log.critical(
            "Executable '%s' was not found! (tool:%s, flow:%s, PATH:%s)",
            e.exec,
            e.tool,
            flow,
            e.path,
        )
        sys.exit(1)
    except FlowSettingsError as e:
        log.critical("%s", e)
        if options.debug:
            raise e from None
        sys.exit(1)


@cli.command(context_settings=CONTEXT_SETTINGS, short_help="List available flows.")
def list_flows():
    table = Table(
        title="Available flows",
        show_header=True,
        header_style="bold yellow",
        title_style=Style(frame=True, bold=True),
        box=box.HEAVY_HEAD,
        show_lines=True,
    )
    table.add_column("Flow", header_style="bold green", style="bold")
    table.add_column("Description")
    table.add_column("Class", style="dim")
    super_flow_doc = inspect.getdoc(Flow)
    for cls_name, (_mod, cls) in sorted(registered_flows.items()):
        doc = inspect.getdoc(cls)
        if doc == super_flow_doc:
            doc = "<no description>"
        table.add_row(
            cls_name,
            doc,
            str(cls.__module__) + "." + cls.__name__,
        )
    console.print(table)


def print_flow_settings(flow, options: XedaOptions):
    flow_class = discover_flow_class(flow)
    schema = flow_class.Settings.schema(by_alias=True)
    type_defs = {}

    def get_type(field):
        typ = field.get("type")
        ref = field.get("$ref")
        if typ is None and ref:
            typ = ref.split("/")[-1]
            typ_def = schema.get("definitions", {}).get(typ)
            if typ not in type_defs:
                type_defs[typ] = typ_def
            elif type_defs[typ] != typ_def:
                log.critical(
                    "type definition for %s changed!\nPrevious def:\n %s new def:\n %s",
                    typ,
                    type_defs[typ],
                    typ_def,
                )
            return Text(typ, style="blue")
        additional = field.get("additionalProperties")
        if typ == "object" and additional:
            return Text(f"Dict[string -> {additional.get('type')}]")

        def join_types(lst, joiner, style="red"):
            return reduce(
                lambda x, y: x + Text(joiner, style) + y, (get_type(t) for t in lst)
            )

        allof = field.get("allOf")
        if allof:
            return join_types(allof, "+")
        anyOf = field.get("anyOf")
        if anyOf:
            return join_types(anyOf, " or ")
        return Text(typ)

    table = Table(
        title=f"{flow} settings",
        show_header=True,
        header_style="bold yellow",
        title_style=Style(frame=True, bold=True),
        box=box.HEAVY_HEAD,
        show_lines=True,
    )
    table.add_column("Property", header_style="bold green", style="bold")
    table.add_column("Type", max_width=32)
    table.add_column("Default", max_width=42)
    table.add_column("Description")

    for name, field in schema.get("properties", {}).items():
        if name in ["results"] or name in Flow.Settings.__fields__.keys():
            continue
        required = name in schema.get("required", [])
        desc: str = field.get("description", "")
        typ = get_type(field)
        default = field.get("default")
        req_or_def = "[red]<required>[/red]" if required else str(default)
        table.add_row(name, typ, req_or_def, desc)
    console.print(table)

    for typ_name, typ_def in list(type_defs.items()):
        for prop, prop_def in typ_def.get("properties").items():
            get_type(prop_def)
    for typ_name, typ_def in list(type_defs.items()):
        c = "blue"
        table = Table(
            title=f"Type [{c}]{typ_name}[/{c}]",
            show_header=True,
            header_style="bold green",
            title_style=Style(frame=True, bold=True),
            box=box.SQUARE_DOUBLE_HEAD,
            show_lines=True,
        )
        if options.debug:
            console.print(f"Type: {typ_name}")
            console.print_json(data=typ_def)
        table.add_column("property")
        table.add_column("type")
        table.add_column("description")
        for prop, prop_def in typ_def.get("properties").items():
            if prop_def.get("hidden_from_schema"):
                continue
            if (
                typ_name.endswith("__Settings")
                and prop in Flow.Settings.__fields__.keys()
            ):
                continue
            desc = prop_def.get("description", prop_def.get("title", "-"))
            desc = re.sub(r"\s*\.*\s*$", "", desc)
            table.add_row(prop, get_type(prop_def), desc)
        console.print(table)


@cli.command(
    context_settings=CONTEXT_SETTINGS, short_help="List flow settings information"
)
@click.argument(
    "flow",
    metavar="FLOW_NAME",
    required=True,
)
@click.pass_context
def list_settings(ctx: click.Context, flow):
    print_flow_settings(flow, options=ctx.obj)


@cli.command(
    context_settings=CONTEXT_SETTINGS,
    short_help="Design-space exploration: run several instances of a flow to find optimal parameters and results",
)
@click.argument("flow")
@click.option(
    "--max-cpus",
    default=max(1, multiprocessing.cpu_count()),
    type=int,
    help="Maximum total number of logical CPU cores to use.",
    show_default=True,
    show_envvar=True,
)
@click.pass_context
def dse(ctx, flow, max_cpus):  # pylint: disable=unused-argument
    """Design-space exploration (e.g. fmax)"""


SHELLS = {
    "bash": {
        "eval_file": "~/.bashrc",
        "eval": 'eval "$(_XEDA_COMPLETE=bash_source xeda)"',
        "completion_file": None,
    },
    "zsh": {
        "eval_file": "~/.zshrc",
        "eval": 'eval "$(_XEDA_COMPLETE=zsh_source xeda)"',
        "completion_file": None,
    },
    "fish": {
        "eval_file": "~/.config/fish/completions/xeda.fish",
        "eval": "eval (env _XEDA_COMPLETE=fish_source xeda)",
        "completion_file": None,
    },
}


@cli.command(
    context_settings=CONTEXT_SETTINGS,
    short_help="Shell completion",
    help=f"Name of the shell. Supported shells are: {', '.join(SHELLS.keys())}",
)
@click.argument(
    "shell",
    required=False,
    type=click.Choice(list(SHELLS.keys()), case_sensitive=False),
)
@click.option(
    "--stdout",
    is_flag=True,
    cls=ClickMutex,
    help="""Produce the shell completion script and output it to the standard output.\n
    Example usage:\n
         $ xeda completion zsh --stdout  > $HOME/.zsh/completion/_xeda\n
         $ xeda completion bash --stdout > $HOME/.local/share/bash-completion/xeda
    """,
)
@click.pass_context
def completion(_ctx: click.Context, stdout, shell=None):
    """Xeda shell auto-completion"""
    os_default_shell_name = None
    os_default_shell = os.environ.get("SHELL")
    if os_default_shell:
        os_default_shell_name = os_default_shell.split(os.sep)[-1]
    if os_default_shell_name:
        if not shell:
            shell = os_default_shell_name
        elif os_default_shell_name != shell and not stdout:
            console.print(
                f"[yellow]WARNING:[/] Current default shell ([bold]{os_default_shell}[/]) is different from the specified shell [bold]{shell}[/b]"
            )
    if stdout:
        completion_class = get_completion_class(shell)
        if completion_class:
            complete = completion_class(
                cli=cli, ctx_args={}, prog_name=__package__, complete_var="source_xeda"
            )
            print(complete.source())
    else:
        console.print(
            f"""
    Make sure 'xeda' executable is properly installed and is accessible through the shell's PATH.
        [yellow]$ xeda --version[/]
    Add the following line to [magenta underline]{SHELLS[shell]['eval_file']}[/]:
        {SHELLS[shell]['eval']}
            """,
            highlight=False,
        )


@cli.command()
@click.argument("subcommand", required=False)
@click.pass_context
def help(ctx: click.Context, subcommand=None):  # pylint: disable=redefined-builtin
    if subcommand:
        subcommand_obj = cli.get_command(ctx, subcommand)
        if subcommand_obj is None:
            click.echo("I don't know that command.")
        else:
            click.echo(subcommand_obj.get_help(ctx))
    else:
        click.echo(cli.get_usage(ctx))



"""Flow runner"""
import hashlib
import importlib
import json
import logging
import os
import time
from datetime import datetime, timedelta
from pathlib import Path, PosixPath
from typing import Any, Dict, Mapping, Optional, Set, Tuple, Type, Union

# fmt: off
from pathvalidate import sanitize_filename  # type: ignore # pyright: reportPrivateImportUsage=none

from rich import box, print_json
from rich.style import Style
from rich.table import Table
from rich.text import Text

from ..console import console
from ..dataclass import asdict
from ..design import Design
from ..flows.flow import Flow, FlowDependencyFailure, registered_flows
from ..tool import NonZeroExitCode
from ..utils import WorkingDirectory, backup_existing, dump_json, snakecase_to_camelcase
from ..version import __version__

__all__ = [
    "get_flow_class",
    "FlowNotFoundError",
    "FlowRunner",
    "DefaultRunner",
    "print_results",
]

log = logging.getLogger(__name__)


def print_results(
    flow: Flow,
    results: Optional[Dict[str, Any]] = None,
    title: Optional[str] = None,
    subset: Optional[Set[str]] = None,
    skip_if_empty: Optional[Set[str]] = None,
) -> None:
    if results is None:
        results = flow.results
    console.print()
    table = Table(
        title=title,
        title_style=Style(frame=True, bold=True),
        show_header=False,
        box=box.ROUNDED,
        show_lines=True,
    )
    table.add_column(style="bold", no_wrap=True)
    table.add_column(justify="right")
    for k, v in results.items():
        if skip_if_empty and k in skip_if_empty and not v:
            continue
        if v is not None and not k.startswith("_"):
            if k == "success":
                text = "OK :heavy_check_mark-text:" if v else "FAILED :cross_mark-text:"
                color = "green" if v else "red"
                table.add_row("Status", text, style=Style(color=color))
                continue
            if subset and k not in subset:
                continue
            if k == "design":
                table.add_row("Design Name", Text(v), style=Style(dim=True))
                continue
            if k == "flow":
                table.add_row("Flow Name", Text(v), style=Style(dim=True))
                continue
            if k == "runtime":
                table.add_row(
                    "Running Time",
                    str(timedelta(seconds=round(v))),
                    style=Style(dim=True),
                )
                continue
            if isinstance(v, float):
                v = f"{v:.3f}"
            table.add_row(k, str(v))
    console.print(table)


class FlowNotFoundError(Exception):
    pass


def get_flow_class(
    flow_name: str, module_name: str = "xeda.flows", package: str = __package__
) -> Type[Flow]:
    _mod, flow_class = registered_flows.get(flow_name, (None, None))
    if flow_class is None:
        log.warning(
            "Flow %s was not found in registered flows. Trying to load using importlib.import_module",
            flow_name,
        )
        try:
            module = importlib.import_module(module_name)
        except ModuleNotFoundError as e:
            raise FlowNotFoundError() from e
        flow_class_name = snakecase_to_camelcase(flow_name)
        if module:
            try:
                flow_class = getattr(module, flow_class_name)
            except AttributeError:
                pass
        if not flow_class or not issubclass(flow_class, Flow):
            raise FlowNotFoundError()
    return flow_class


def _semantic_hash(data: Any, debug=False) -> str:
    def _sorted_dict_str(data: Any) -> Any:
        if isinstance(data, (dict, Mapping)):
            return {k: _sorted_dict_str(data[k]) for k in sorted(data.keys())}
        if isinstance(data, list):
            return [_sorted_dict_str(val) for val in data]
        if hasattr(data, "__dict__"):
            return _sorted_dict_str(data.__dict__)
        return str(data)

    def _get_digest(b: bytes) -> str:
        return hashlib.sha256(b).hexdigest()[:16]

    r = repr(_sorted_dict_str(data))
    if debug:
        print("\nSEMANTIC_HASH:", r)
    return _get_digest(bytes(r, "UTF-8"))


class FlowRunner:
    """
    Manage running flows and their dependencies.
    1. Instantiate instance of flow class with proper settings assigned (__init__)
    2. call Flow.init()
    3. Run all dependency flows (asked by the flow, during Flow.init)
    3. Run the flow by calling run()
    4. Run flow's parse_reports() ## TODO parse_reports will be renamed
    5. Evaluate and print the results
    """

    def __init__(
        self,
        xeda_run_dir: Union[str, os.PathLike] = "xeda_run",
        debug: bool = False,
        dump_settings_json: bool = True,
        display_results: bool = True,
        dump_results_json: bool = True,
        cached_dependencies: bool = False,  # do not run dependencies if previous run results exist. Uses flow run_dir names including design and flow.settings hashes
        run_in_existing_dir: bool = False,  # DO NOT USE! Only for development!
    ) -> None:
        if debug:
            logging.getLogger().setLevel(logging.DEBUG)
            log.root.setLevel(logging.DEBUG)
        self.debug = debug
        log.debug("%s xeda_run_dir=%s", self.__class__.__name__, xeda_run_dir)
        xeda_run_dir = Path(xeda_run_dir).resolve()
        xeda_run_dir.mkdir(exist_ok=True, parents=True)
        self.xeda_run_dir: Path = xeda_run_dir
        self.cached_dependencies: bool = cached_dependencies
        self.display_results: bool = display_results
        self.dump_results_json: bool = dump_results_json
        self.dump_settings_json: bool = dump_settings_json
        self.run_in_existing_dir: bool = run_in_existing_dir

    def _get_flow_run_path(
        self,
        design_name: str,
        flow_name: str,
        design_hash: Optional[str] = None,
        flowrun_hash: Optional[str] = None,
    ) -> Path:
        design_subdir = design_name
        flow_subdir = flow_name
        if self.cached_dependencies:
            if design_hash:
                design_subdir += f"_{design_hash}"
            if flowrun_hash:
                flow_subdir += f"_{flowrun_hash}"

        run_path: Path = (
            self.xeda_run_dir / sanitize_filename(design_subdir) / flow_subdir
        )
        return run_path

    def run_flow(
        self,
        flow_class: Union[str, Type[Flow]],
        design: Design,
        flow_settings: Union[None, Dict[str, Any], Flow.Settings] = None,
    ) -> Flow:
        return self._run_flow(flow_class, design, flow_settings, None)[-1]

    def _run_flow(
        self,
        flow_class: Union[str, Type[Flow]],
        design: Design,
        flow_settings: Union[None, Dict[str, Any], Flow.Settings],
        depender: Optional[Flow],
    ) -> Tuple[Flow, ...]:
        if self.run_in_existing_dir:
            log.error(
                "run_in_existing_dir should only be used during Xeda's development!"
            )
        if isinstance(flow_class, str):
            flow_class = get_flow_class(flow_class)
        if flow_settings is None:
            flow_settings = {}
        elif isinstance(flow_settings, Flow.Settings):
            flow_settings = asdict(flow_settings)
        flow_settings = flow_class.Settings(**flow_settings)
        if self.debug:
            flow_settings.debug = True

        flow_name = flow_class.name
        # GOTCHA: design contains tb settings even for simulation flows
        # removing tb from hash for sim flows creates a mismatch for different flows of the same design
        design_hash = _semantic_hash(design)
        flowrun_hash = _semantic_hash(
            dict(
                flow_name=flow_name,
                flow_settings=flow_settings,
                xeda_version=__version__,
            ),
        )
        run_path = self._get_flow_run_path(
            design.name,
            flow_name,
            design_hash,
            flowrun_hash,
        )

        settings_json = run_path / "settings.json"
        results_json = run_path / "results.json"

        previous_results = None
        if (
            depender
            and self.cached_dependencies
            and run_path.exists()
            and settings_json.exists()
            and results_json.exists()
        ):
            prev_results, prev_settings = None, None
            try:
                with open(settings_json) as f:
                    prev_settings = json.load(f)
                with open(results_json) as f:
                    prev_results = json.load(f)
            except TypeError:
                pass
            except ValueError:
                pass
            if prev_results and prev_results.get("success") and prev_settings:
                if (
                    prev_settings.get("flow_name") == flow_name
                    and prev_settings.get("design_hash") == design_hash
                    and prev_settings.get("flowrun_hash") == flowrun_hash
                ):
                    previous_results = prev_results
                else:
                    log.warning(
                        "%s does not contain the expected settings", prev_settings
                    )
            else:
                log.warning(
                    "Could not find valid results/settings from a previous run in %s",
                    run_path,
                )

        if not previous_results:
            if not self.run_in_existing_dir and run_path.exists():
                backup_existing(run_path)
            run_path.mkdir(parents=True)

            if self.dump_settings_json:
                log.info("dumping effective settings to %s", settings_json)
                all_settings = dict(
                    design=design,
                    design_hash=design_hash,
                    flow_name=flow_name,
                    flow_settings=flow_settings,
                    xeda_version=__version__,
                    flowrun_hash=flowrun_hash,
                )
                dump_json(all_settings, settings_json)

        with WorkingDirectory(run_path):
            log.debug("Instantiating flow from %s", flow_class)
            flow = flow_class(flow_settings, design, run_path)
            flow.design_hash = design_hash
            flow.flow_hash = flowrun_hash
            flow.timestamp = datetime.now().strftime("%Y-%m-%d-%H%M%S")
            # flow execution time includes init() as well as execution of all its dependency flows
            flow.init_time = time.monotonic()
            flow.init()

        for dep_cls, dep_settings in flow.dependencies:
            # merge with existing self.flows[dep].settings
            # NOTE this allows dependency flow to make changes to 'design'
            log.info(
                "Running dependency: %s (%s.%s)",
                dep_cls.name,
                dep_cls.__module__,
                dep_cls.__qualname__,
            )
            completed_dep = self._run_flow(dep_cls, design, dep_settings, depender=flow)
            if not completed_dep:
                log.critical("Dependency flow: %s failed!", dep_cls.name)
                raise FlowDependencyFailure()  # TODO
            flow.completed_dependencies.extend(completed_dep)

        flow.results["design"] = flow.design.name
        flow.results["flow"] = flow.name
        success = True

        if previous_results:
            log.warning("Using previous run results and artifacts from %s", run_path)
            flow.results = previous_results
        else:
            with WorkingDirectory(run_path):
                try:
                    flow.run()
                except NonZeroExitCode as e:
                    log.critical(
                        "Execution of %s returned %d", e.command_args[0], e.exit_code
                    )
                    success = False
                if flow.init_time is not None:
                    flow.results.runtime = time.monotonic() - flow.init_time
                try:
                    success &= flow.parse_reports()
                except Exception as e:  # pylint: disable=broad-except
                    log.critical("parse_reports throw an exception: %s", e)
                    if success:  # if so far so good this is a bug!
                        raise e from None
                    success = False
                if not success:
                    log.error("Failure was reported in the parsed results.")
                flow.results.success = success

        if flow.artifacts:
            def default_encoder(x: Any) -> str:
                if isinstance(x, (PosixPath, os.PathLike)):
                    return str(os.path.relpath(x, flow.run_path))
                return str(x)

            print(f"Generated artifacts in {flow.run_path}:")  # FIXME
            print_json(data=flow.artifacts, default=default_encoder)  # FIXME

        flow.results.success = success
        if not success:
            # set success=false if execution failed
            log.critical("%s failed!", flow.name)

        if self.dump_results_json:
            dump_json(flow.results, results_json)
            log.info("Results written to %s", results_json)

        if self.display_results:
            print_results(
                flow,
                title=f"{flow.name} Results",
                skip_if_empty={"artifacts", "reports"},
            )
        return (*flow.completed_dependencies, flow)


class DefaultRunner(FlowRunner):
    """Executes a flow and its dependencies and then reports selected results"""



import logging
from typing import Any, Dict

from ..flow import FpgaSynthFlow
from ...dataclass import validator
from . import Vivado
from .vivado_synth import RunOptions, VivadoSynth

log = logging.getLogger(__name__)

# curated options based on experiments and:
# see https://www.xilinx.com/support/documentation/sw_manuals/xilinx2020_1/ug904-vivado-implementation.pdf
# and https://www.xilinx.com/support/documentation/sw_manuals/xilinx2020_1/ug901-vivado-synthesis.pdf
xeda_strategies: Dict[str, Dict[str, Any]] = {
    "Debug": {
        "synth": {
            "-assert": None,
            "-debug_log": None,
            "-flatten_hierarchy": "none",
            "-keep_equivalent_registers": None,
            "-no_lc": None,
            "-fsm_extraction": "off",
            "-directive": "RuntimeOptimized",
        },
        "opt": {"-directive": "RuntimeOptimized"},
        "place": {"-directive": "RuntimeOptimized"},
        "place_opt": {},
        "route": {"-directive": "RuntimeOptimized"},
        "phys_opt": {"-directive": "RuntimeOptimized"},
    },
    "Runtime": {
        "synth": {"-directive": "RuntimeOptimized"},
        "opt": {"-directive": "RuntimeOptimized"},
        "place": {"-directive": "RuntimeOptimized"},
        "place_opt": {},
        # with -ultrathreads results are not reproducible!
        # OR "-no_timing_driven -ultrathreads",
        "route": {"-directive": "RuntimeOptimized"},
        "phys_opt": {"-directive": "RuntimeOptimized"},
    },
    "Default": {
        "synth": {"-flatten_hierarchy": "rebuilt", "-directive": "Default"},
        "opt": {"-directive": "ExploreWithRemap"},
        "place": {"-directive": "Default"},
        "place_opt": {},
        "route": {"-directive": "Default"},
        "phys_opt": {"-directive": "Default"},
    },
    "Timing": {
        # -mode: default, out_of_context
        # -flatten_hierarchy: rebuilt, full; equivalent in terms of QoR?
        # -no_lc: When checked, this option turns off LUT combining
        # -keep_equivalent_registers -no_lc
        "synth": {
            "-flatten_hierarchy": "rebuilt",
            "-retiming": None,
            "-directive": "PerformanceOptimized",
            "-fsm_extraction": "one_hot",
            #   "-resource_sharing off",
            #   "-no_lc",
            "-shreg_min_size": "5",
            #   "-keep_equivalent_registers "
        },
        "opt": {"-directive": "ExploreWithRemap"},
        "place": {"-directive": "ExtraPostPlacementOpt"},
        "place_opt": {
            "-retarget": None,
            "-propconst": None,
            "-sweep": None,
            "-aggressive_remap": None,
            "-shift_register_opt": None,
        },
        "phys_opt": {"-directive": "AggressiveExplore"},
        "place_opt2": {"-directive": "Explore"},
        # "route": "-directive NoTimingRelaxation",
        "route": {"-directive": "AggressiveExplore"},
    },
    # "ExtraTimingCongestion": {
    #     "synth": ["-flatten_hierarchy": "full",
    #               "-retiming",
    #               "-directive": "PerformanceOptimized",
    #               "-fsm_extraction": "one_hot",
    #               "-resource_sharing off",
    #               "-shreg_min_size 10",
    #               "-keep_equivalent_registers",
    #               ],
    #     "opt": ["-directive ExploreWithRemap"],
    #     "place": ["-directive AltSpreadLogic_high"],
    #     "place_opt": ['-retarget', '-propconst', '-sweep', '-remap', '-muxf_remap', '-aggressive_remap', '-shift_register_opt'],
    #     "place_opt2": ["-directive Explore"],
    #     "phys_opt": ["-directive AggressiveExplore"],
    #     "route": ["-directive AlternateCLBRouting"],
    # },
    "ExtraTiming": {
        "synth": {
            "-flatten_hierarchy": "full",
            "-retiming": None,
            "-directive": "PerformanceOptimized",
            "-fsm_extraction": "one_hot",
            "-resource_sharing": "off",
            #   "-no_lc",
            "-shreg_min_size": "10",
            "-keep_equivalent_registers": None,
        },
        "opt": {"-directive": "ExploreWithRemap"},
        "place": {"-directive": "ExtraTimingOpt"},
        "place_opt": {
            "-retarget": None,
            "-propconst": None,
            "-sweep": None,
            "-muxf_remap": None,
            "-aggressive_remap": None,
            "-shift_register_opt": None,
        },
        "place_opt2": {"-directive": "Explore"},
        "phys_opt": {"-directive": "AggressiveExplore"},
        "route": {"-directive": "NoTimingRelaxation"},
    },
    # "ExtraTimingAltRouting": {
    #     # -mode: default, out_of_context
    #     # -flatten_hierarchy: rebuilt, full; equivalent in terms of QoR?
    #     # -no_lc: When checked, this option turns off LUT combining
    #     # -keep_equivalent_registers -no_lc
    #     "synth": ["-flatten_hierarchy full",
    #               "-retiming",
    #               "-directive PerformanceOptimized",
    #               "-fsm_extraction one_hot",
    #               #   "-resource_sharing off",
    #               #   "-no_lc",
    #               "-shreg_min_size 5",
    #               "-keep_equivalent_registers "
    #               ],
    #     "opt": ["-directive ExploreWithRemap"],
    #     "place": ["-directive ExtraTimingOpt"],
    #     "place_opt": ['-retarget', '-propconst', '-sweep', '-aggressive_remap', '-shift_register_opt'],
    #     "phys_opt": ["-directive AggressiveExplore"],
    #     # "route": "-directive NoTimingRelaxation",
    #     "route": ["-directive AlternateCLBRouting"],
    # },
    # "Area": {
    #     # AreaOptimized_medium or _high prints error messages in Vivado 2020.1: "unexpected non-zero reference counts", but succeeeds and post-impl sim is OK too
    #     "synth": ["-flatten_hierarchy full", "-control_set_opt_threshold 1", "-shreg_min_size 3", "-resource_sharing auto", "-directive AreaOptimized_medium"],
    #     # if no directive: -resynth_seq_area
    #     "opt": "-directive ExploreArea",
    #     "place": "-directive Default",
    #     "place_opt": "-directive ExploreArea",
    #     # "place_opt": ['-retarget', '-propconst', '-sweep', '-aggressive_remap', '-shift_register_opt',
    #     #               '-dsp_register_opt', '-bram_power_opt', '-resynth_seq_area', '-merge_equivalent_drivers'],
    #     # if no directive: -placement_opt
    #     "phys_opt": "-directive Explore",
    #     "route": "-directive Explore",
    # },
    # "AreaHigh": {
    #     # AreaOptimized_medium or _high prints error messages in Vivado 2020.1: "unexpected non-zero reference counts", but succeeeds and post-impl sim is OK too
    #     "synth": ["-flatten_hierarchy full", "-control_set_opt_threshold 1", "-shreg_min_size 3", "-resource_sharing on", "-directive AreaOptimized_high"],
    #     # if no directive: -resynth_seq_area
    #     "opt": "-directive ExploreArea",
    #     "place": "-directive Default",
    #     "place_opt": "-directive ExploreArea",
    #     # "place_opt": ['-retarget', '-propconst', '-sweep', '-aggressive_remap', '-shift_register_opt',
    #     #               '-dsp_register_opt', '-bram_power_opt', '-resynth_seq_area', '-merge_equivalent_drivers'],
    #     # if no directive: -placement_opt
    #     "phys_opt": "-directive Explore",
    #     "route": "-directive Explore",
    # },
    # "AreaPower": {
    #     # AreaOptimized_medium or _high prints error messages in Vivado 2020.1: "unexpected non-zero reference counts", but succeeeds and post-impl sim is OK too
    #     "synth": ["-flatten_hierarchy full", "-control_set_opt_threshold 1", "-shreg_min_size 3", "-resource_sharing auto", "-gated_clock_conversion auto", "-directive AreaOptimized_medium"],
    #     "opt": ["-directive ExploreArea"],
    #     "place": "-directive Default",
    #     "place_opt": ['-retarget', '-propconst', '-sweep', '-aggressive_remap', '-shift_register_opt', '-dsp_register_opt', '-resynth_seq_area', '-merge_equivalent_drivers'],
    #     "place_opt2": ["-directive ExploreArea"],
    #     # FIXME!!! This is the only option that results in correct post-impl timing sim! Why??!
    #     "phys_opt": ["-directive AggressiveExplore"],
    #     "route": ["-directive Explore"],
    # },
    # "AreaTiming": {
    #     "synth": ["-flatten_hierarchy rebuilt", "-retiming"],
    #     # if no directive: -resynth_seq_area
    #     "opt": ["-directive ExploreWithRemap"],
    #     "place": ["-directive ExtraPostPlacementOpt"],
    #     # "place_opt": ["-directive ExploreArea"],
    #     "place_opt": ['-retarget', '-propconst', '-sweep', '-aggressive_remap', '-shift_register_opt', '-dsp_register_opt', '-resynth_seq_area', '-merge_equivalent_drivers'],
    #     "place_opt2": ["-directive ExploreArea"],
    #     # if no directive: -placement_opt
    #     "phys_opt": "-directive AggressiveExplore",
    #     "route": "-directive Explore",
    # },
    # "AreaExploreWithRemap": {
    #     "synth": ["-flatten_hierarchy full", "-retiming"],
    #     # if no directive: -resynth_seq_area
    #     "opt": "-directive ExploreWithRemap",
    #     "place": "-directive Default",
    #     "place_opt": "-directive ExploreWithRemap",
    #     # "place_opt": ['-retarget', '-propconst', '-sweep', '-aggressive_remap', '-shift_register_opt',
    #     #               '-dsp_register_opt', '-bram_power_opt', '-resynth_seq_area', '-merge_equivalent_drivers'],
    #     # if no directive: -placement_opt
    #     "phys_opt": "-directive Explore",
    #     "route": "-directive Explore",
    # },
    # "AreaExploreWithRemap2": {
    #     "synth": [],
    #     # if no directive: -resynth_seq_area
    #     "opt": "-directive ExploreArea",
    #     "place": "-directive Default",
    #     "place_opt": "-directive ExploreWithRemap",
    #     # "place_opt": ['-retarget', '-propconst', '-sweep', '-aggressive_remap', '-shift_register_opt',
    #     #               '-dsp_register_opt', '-bram_power_opt', '-resynth_seq_area', '-merge_equivalent_drivers'],
    #     # if no directive: -placement_opt
    #     "phys_opt": "-directive Explore",
    #     "route": "-directive Explore",
    # },
    # "AreaExplore": {
    #     "synth": ["-flatten_hierarchy full"],
    #     # if no directive: -resynth_seq_area
    #     "opt": "-directive ExploreArea",
    #     "place": "-directive Default",
    #     "place_opt": "-directive ExploreArea",
    #     # "place_opt": ['-retarget', '-propconst', '-sweep', '-aggressive_remap', '-shift_register_opt',
    #     #               '-dsp_register_opt', '-bram_power_opt', '-resynth_seq_area', '-merge_equivalent_drivers'],
    #     # if no directive: -placement_opt
    #     "phys_opt": "-directive Explore",
    #     "route": "-directive Explore",
    # },
    # "Power": {
    #     "synth": ["-flatten_hierarchy full", "-gated_clock_conversion auto", "-control_set_opt_threshold 1", "-shreg_min_size 3", "-resource_sharing auto"],
    #     # if no directive: -resynth_seq_area
    #     "opt": "-directive ExploreSequentialArea",
    #     "place": "-directive Default",
    #     "place_opt": "-directive ExploreSequentialArea",
    #     # ['-retarget', '-propconst', '-sweep', '-aggressive_remap', '-shift_register_opt',
    #     #   '-dsp_register_opt', '-bram_power_opt', '-resynth_seq_area', '-merge_equivalent_drivers'],
    #     # if no directive: -placement_opt
    #     "phys_opt": "-directive Explore",
    #     "route": "-directive Explore",
    # }
}


def _vivado_steps(strategy: str, s: str):
    if s == "synth":
        steps = ["synth", "opt"]
    else:
        steps = [
            "place",
            "place_opt",
            "place_opt2",
            "phys_opt",
            "phys_opt2",
            "route",
        ]
    return {step: xeda_strategies[strategy].get(step) for step in steps}


class VivadoAltSynth(Vivado, FpgaSynthFlow):
    """Synthesize with Xilinx Vivado using an alternative TCL-based flow"""

    class Settings(VivadoSynth.Settings):
        synth: RunOptions = RunOptions(
            strategy="Default", steps=_vivado_steps("Default", "synth")
        )
        impl: RunOptions = RunOptions(
            strategy="Default", steps=_vivado_steps("Default", "impl")
        )

        # pylint: disable=no-self-argument,no-self-use
        @validator("synth")
        def validate_synth(cls, v: RunOptions):
            if v.strategy and not v.steps:
                v.steps = _vivado_steps(v.strategy, "synth")
            return v

        # pylint: disable=no-self-argument,no-self-use
        @validator("impl", always=True)
        def validate_impl(cls, v: RunOptions, values):
            if not v.strategy:
                v.strategy = values.get("synth", {}).get("strategy")
            if v.strategy and not v.steps:
                v.steps = _vivado_steps(v.strategy, "impl")
            return v

    def run(self):
        clock_xdc_path = self.copy_from_template("clock.xdc")
        ss = self.settings
        assert isinstance(ss, self.Settings)

        if ss.out_of_context:
            if "synth" in ss.synth.steps and ss.synth.steps["synth"] is not None:
                ss.synth.steps["synth"]["-mode"] = "out_of_context"

        blacklisted_resources = ss.blacklisted_resources
        if "bram" in blacklisted_resources and "bram_tile" not in blacklisted_resources:
            blacklisted_resources.append("bram_tile")

        log.info("blacklisted_resources: %s", blacklisted_resources)

        # if 'bram_tile' in blacklisted_resources:
        #     # FIXME also add -max_uram 0 for ultrascale+
        #     settings.synth.steps['synth'].append('-max_bram 0')
        # if 'dsp' in blacklisted_resources:
        #     settings.synth.steps['synth'].append('-max_dsp 0')

        self.add_template_filter(
            "flatten_dict",
            lambda d: " ".join([f"{k} {v}" if v else k for k, v in d.items()]),
        )

        script_path = self.copy_from_template(
            "vivado_alt_synth.tcl",
            xdc_files=[clock_xdc_path],
        )
        self.vivado.run("-source", script_path)



import logging
import os
from pathlib import Path
from typing import Union

from vcd import gtkw

log = logging.getLogger(__name__)


def _add_sig(g: gtkw.GTKWSave, root_group, signals):
    i = 0
    while i < len(signals):
        sig = signals[i]
        if sig[: len(root_group)] == root_group:  # in this group
            if len(sig) - len(root_group) > 1:
                new_group = sig[len(root_group)]
                with g.group(new_group, closed=True):
                    i += _add_sig(g, root_group + [new_group], signals[i:])
            else:
                g.trace(".".join(sig))
                i += 1
        else:
            break
    return i


def gen_gtkw(dump_file: Union[str, os.PathLike], signals, root_group):
    root_group = root_group or []
    save_file = Path(dump_file).with_suffix(".gtkw")
    log.info("generated gtkwave save file: %s", save_file.absolute())
    with open(save_file, "w") as f:
        g = gtkw.GTKWSave(f)
        g.dumpfile(str(dump_file), abspath=not os.path.isabs(dump_file))
        _add_sig(g, root_group, signals)



import os
import tempfile
from pathlib import Path

from xeda import Design
from xeda.flow_runner import DefaultRunner
from xeda.flows import VivadoSynth
from xeda.flows.flow import FPGA

TESTS_DIR = Path(__file__).parent.absolute()
RESOURCES_DIR = TESTS_DIR / "resources"
EXAMPLES_DIR = TESTS_DIR.parent / "examples"


def test_vivado_synth_template() -> None:
    path = RESOURCES_DIR / "design0/design0.toml"
    assert path.exists()
    design = Design.from_toml(RESOURCES_DIR / "design0/design0.toml")
    settings = VivadoSynth.Settings(fpga=FPGA(part="abcd"), clock_period=5.5)  # type: ignore
    run_dir = Path.cwd() / "vivado_synth_run"
    run_dir.mkdir(exist_ok=True)
    flow = VivadoSynth(settings, design, run_dir)  # type: ignore
    tcl_file = flow.copy_from_template(
        "vivado_synth.tcl", xdc_files=[], reports_tcl="reports_tcl"
    )
    with open(run_dir / tcl_file) as f:
        vivado_tcl = f.read()
    expected_lines = [
        "set_property generic {G_IN_WIDTH=32} [current_fileset]",
        "set_property generic {G_ITERATIVE=1'b1} [current_fileset]",
        'set_property generic {G_STR=\\"abcd\\"} [current_fileset]',
        "set_property generic {G_BITVECTOR=7'b0101001} [current_fileset]",
    ]
    for line in expected_lines:
        assert line in vivado_tcl


def test_vivado_synth_py() -> None:
    path = RESOURCES_DIR / "design0/design0.toml"
    # Append to PATH so if the actual tool exists, would take precedences.
    os.environ["PATH"] += os.pathsep + os.path.join(TESTS_DIR, "fake_tools")
    assert path.exists()
    design = Design.from_toml(EXAMPLES_DIR / "vhdl" / "sqrt" / "sqrt.toml")
    settings = dict(fpga=FPGA("xc7a12tcsg325-1"), clock_period=5.5)
    with tempfile.TemporaryDirectory() as run_dir:
        print("Xeda run dir: ", run_dir)
        xeda_runner = DefaultRunner(run_dir, debug=True)
        flow = xeda_runner.run_flow(VivadoSynth, design, settings)
        settings_json = flow.run_path / "settings.json"
        results_json = flow.run_path / "results.json"
        assert settings_json.exists()
        assert results_json.exists()
        assert flow.succeeded
        assert 0.3 < flow.results.runtime  # type: ignore


if __name__ == "__main__":
    test_vivado_synth_py()



"""Utilities for command line interface"""
import logging
import sys
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import click
from click_help_colors import HelpColorsGroup
from overrides import overrides

from .console import console
from .dataclass import asdict
from .flow_runner import FlowNotFoundError, get_flow_class
from .flows.flow import Flow
from .utils import set_hierarchy, try_convert

__all__ = [
    "ClickMutex",
    "OptionEatAll",
    "ConsoleLogo",
    "XedaHelpGroup",
    "discover_flow_class",
]


log = logging.getLogger(__name__)


class ClickMutex(click.Option):
    """Mutual exclusion of options"""

    def __init__(self, *args, **kwargs):
        self.mutually_exclusive_with: list = kwargs.pop("mutually_exclusive_with", [])
        self.required_if: list = kwargs.pop("required_if", [])
        if self.mutually_exclusive_with:
            kwargs["help"] = (
                kwargs.get("help", "")
                + "Option is mutually exclusive with "
                + ", ".join(self.mutually_exclusive_with)
                + "."
            ).strip()
        if self.required_if:
            kwargs["help"] = (
                kwargs.get("help", "")
                + "Option is required only if "
                + " or ".join(self.mutually_exclusive_with)
                + " is specified "
            ).strip()
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        if self.name in opts:
            for mutex_opt in self.mutually_exclusive_with:
                if mutex_opt in opts:
                    raise click.UsageError(
                        f"Option {self.name} is mutually exclusive with {mutex_opt}."
                    )
        else:
            for dependent_opt in self.required_if:
                if dependent_opt in opts:
                    raise click.UsageError(
                        f"Option {self.name} is required when {dependent_opt} is specified."
                    )

        return super().handle_parse_result(ctx, opts, args)


class OptionEatAll(click.Option):
    """
    Taken from https://stackoverflow.com/questions/48391777/nargs-equivalent-for-options-in-click#answer-48394004.
    """  # noqa: ignore=E501 pylint: disable=C0301

    @overrides
    def __init__(self, *args, **kwargs):
        self.save_other_options = kwargs.pop("save_other_options", True)
        super().__init__(*args, **kwargs)
        self._previous_parser_process: Optional[
            Callable[[Union[str, Tuple[str, ...]], click.parser.ParsingState], None]
        ] = None
        self._eat_all_parser = None

    @overrides
    def add_to_parser(self, parser, ctx):
        def parser_process(value: str, state: click.parser.ParsingState):
            """method to hook to the parser.process"""
            value_list = [value]
            if self._eat_all_parser is not None and self.save_other_options:
                # grab everything up to the next option
                done = False
                while state.rargs:
                    for prefix in self._eat_all_parser.prefixes:
                        if state.rargs[0].startswith(prefix):
                            done = True
                            break
                    if done:
                        break
                    value_list.append(state.rargs.pop(0))
            else:
                # grab everything remaining
                value_list += state.rargs
                state.rargs[:] = []

            # call the actual process
            if self._previous_parser_process is not None:
                self._previous_parser_process(tuple(value_list), state)

        super().add_to_parser(parser, ctx)
        for name in self.opts:
            # pylint: disable=protected-access
            our_parser = parser._long_opt.get(name) or parser._short_opt.get(name)
            if our_parser:
                self._eat_all_parser = our_parser
                self._previous_parser_process = our_parser.process  # type: ignore
                our_parser.process = parser_process
                break


class ConsoleLogo:
    """Xeda ASCII logo on console"""

    logo: str = """
:==-+-:        :=-.-='.---==:::::::==-. .--=:::::::::=.          ,+==::::::+=-.
 .= {X}O{O} -=     .=--.==. *  ,...........:: *:  {X}........{O}  `=-      ,='            *
   :=:--=-  =-..-=.   +  .-:::::::::::' +: {X}=+++++++++-{O} `:+   ,=' .*########:  *
     -=.-:=-.-:=:     +  +:             +: {X}=+++++++++*.{O} `=- .=: :*########%:  *
       -==-.-=-       +  +: ,:::::::-,  +: {X}=++++++++++*:{O} := =: :*%########%:  *
      .==-.-==-       +  +: `:::::::='  +: {X}=++++++++++*:{O} := =: `*%#######%*'  *
     -= -:==. ==-     +  +:             +: {X}=++++++++++*:{O} .= =:                *
   := -==-  ==. ==-   *  `+..........-. +. {X}=+++++++++*:{O} .+  =: .==::::::::=.  *
 .:-.==:     .=- {X}O{O} =. :=+-............: *, {X}:++++++++'{O} .+-   =: -=          +  *
:=--==.        :=--.=: `=+:::::::::::-' `*-::::::::::+-'    `=-='          `-='
    """

    @classmethod
    def print(cls):
        """print the logo if console is wide enough"""
        if console.width >= 80:
            console.print(
                cls.logo.format(X="[dark_orange]", O="[/]"),
                highlight=False,
                emoji=False,
            )


class XedaHelpGroup(HelpColorsGroup):
    """How to display CLI help"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.help_headers_color = "yellow"
        self.help_options_color = "green"

    def format_usage(self, ctx: click.Context, formatter: click.HelpFormatter):
        ConsoleLogo.print()
        super().format_usage(ctx, formatter)


# DictStrHier = Dict[str, "StrOrDictStrHier"]
DictStrHier = Dict[str, Any]
StrOrDictStrHier = Union[str, DictStrHier]


def settings_to_dict(
    settings: Union[
        None, List[str], Tuple[str, ...], Dict[str, StrOrDictStrHier], Flow.Settings
    ],
) -> Dict[str, Any]:
    if not settings:
        return {}
    if isinstance(settings, (tuple, list)):
        res: DictStrHier = {}
        for override in settings:
            sp = override.split("=")
            if len(sp) != 2:
                raise ValueError("Settings should be in KEY=VALUE format!")
            key, val = sp
            set_hierarchy(res, key, try_convert(val, convert_lists=True))
        return res
    if isinstance(settings, Flow.Settings):
        return asdict(settings)
    if isinstance(settings, dict):
        return settings
    raise TypeError(f"overrides is of unsupported type: {type(settings)}")


# xdg_home = os.environ.get('XDG_DATA_HOME', os.path.join(
#     os.environ.get('HOME', str(Path.home())), '.local', 'share'))
# completion_user_dir = os.environ.get(
#     'BASH_COMPLETION_USER_DIR', os.path.join(xdg_home, 'bash-completion'))
# dir = Path(completion_user_dir)
# if not dir.exists():
#     dir.mkdir(parents=True)
# completion_file = dir / 'xeda'
# with open(completion_file, 'w') as f:
#     f.write(completion)
# eager_completion = Path.home() / '.bash_completion'
# source_line = f'. {completion_file} # added by xeda'
# if eager_completion.exists():
#     with open(eager_completion, "r+") as f:
#         for line in f:
#             if source_line in line:
#                 break
#         else:  # else-for "completion clause"
#             f.write(source_line + os.linesep)
# else:
#     with open(eager_completion, "w") as f:
#         f.write(source_line + os.linesep)


def discover_flow_class(flow: str) -> Type[Flow]:
    try:
        return get_flow_class(flow, "xeda.flows", __package__)
    except FlowNotFoundError:
        log.critical(
            "Flow %s is not known to Xeda. Please make sure the name is correctly specified.",
            flow,
        )
        sys.exit(1)



import html
import logging
from typing import Any, Dict
from xml.etree import ElementTree

from .vivado_postsynthsim import VivadoPostsynthSim
from .vivado_alt_synth import VivadoSynth

logger = logging.getLogger(__name__)


class VivadoPower(VivadoPostsynthSim):
    class Settings(VivadoPostsynthSim.Settings):
        saif: str = "activity.saif"
        elab_debug = "typical"
        timing_sim = True
        power_report_xml: str = "power_impl_timing.xml"

    def run(self) -> None:
        super().run()

        assert isinstance(self.settings, self.Settings)
        dep_synth_flow = self.pop_dependency(VivadoSynth)
        assert self.design.tb
        # assert isinstance(dep_synth_flow.settings, VivadoSynth.Settings)
        top = self.design.rtl.top
        script_path = self.copy_from_template(
            "vivado_power.tcl",
            checkpoint=str(
                dep_synth_flow.run_path / f"{top}.runs" / "impl_1" / f"{top}_routed.dcp"
            ),
        )

        self.vivado.run("-source", script_path)

    def parse_power_report(self, report_xml) -> Dict[str, Any]:
        tree = ElementTree.parse(report_xml)
        results = {}
        for tablerow in tree.findall("./section[@title='Summary']/table/tablerow"):
            tablecells = tablerow.findall("tablecell")
            key, value = (
                html.unescape(x.attrib["contents"]).strip() for x in tablecells
            )
            results[key] = value

        for tablerow in tree.findall(
            "./section[@title='Summary']/section[@title='On-Chip Components']/table/tablerow"
        ):
            tablecells = tablerow.findall("tablecell")
            if len(tablecells) >= 2:
                contents = [
                    html.unescape(x.attrib["contents"]).strip() for x in tablecells
                ]
                key = contents[0]
                value = contents[1]
                results[f"Component Power: {key}"] = value

        return results

    def parse_reports(self) -> bool:
        assert isinstance(self.settings, self.Settings)
        report_xml = self.run_path / self.settings.power_report_xml
        results = self.parse_power_report(report_xml)
        self.results.update(**results)
        return True



import logging
from pathlib import Path
from typing import Optional

from ..board import WithFpgaBoardSettings, get_board_data
from ..design import Design
from ..tool import Tool
from .flow import FpgaSynthFlow
from .nextpnr import Nextpnr

__all__ = ["Openfpgaloader"]

log = logging.getLogger(__name__)


class Openfpgaloader(FpgaSynthFlow):
    ofpga_loader = Tool("openFPGALoader")

    class Settings(WithFpgaBoardSettings):
        clock_period: float
        reset: bool = False
        cable: Optional[str] = None

    def __init__(self, flow_settings: Settings, design: Design, run_path: Path):
        super().__init__(flow_settings, design, run_path)
        self.packer: Optional[Tool] = None

    def init(self) -> None:
        assert isinstance(self.settings, self.Settings)
        ss = self.settings
        assert ss.fpga
        self.add_dependency(
            Nextpnr,
            Nextpnr.Settings(
                fpga=ss.fpga, board=ss.board, clock_period=ss.clock_period
            ),  # type: ignore
        )
        if ss.fpga.family == "ecp5":  # FIXME from fpga/board
            self.packer = Tool("ecppack")

    def run(self) -> None:
        assert isinstance(self.settings, self.Settings)
        ss = self.settings
        board_id = ss.board
        board_name = None
        if board_id:
            board_data = get_board_data(board_id)
            if board_data:
                board_name = board_data.get("name")
        next_pnr = self.completed_dependencies[0]
        assert isinstance(next_pnr, Nextpnr)
        assert isinstance(next_pnr.settings, Nextpnr.Settings)
        assert next_pnr.settings.textcfg
        text_cfg = next_pnr.run_path / next_pnr.settings.textcfg
        assert text_cfg.exists(), f"Can't find {text_cfg} generated by Nextpnr!"
        bitstream = f"{board_name}.bit" if board_name else "bitstream.bit"
        if self.packer:
            self.packer.run(text_cfg, bitstream)
        args = ["--bitstream", bitstream]
        if ss.cable:
            args.extend(["--cable", ss.cable])
        elif board_name:
            args.extend(["--board", board_name])
        if ss.fpga.part:
            args.extend(["--fpga-part", ss.fpga.part])
        if ss.reset:
            args.append("--reset")
        if ss.verbose:
            args.append("--verbose")
        self.ofpga_loader.run(*args)



"""test Intel Quartus flow"""
import os
import tempfile
import logging
from pathlib import Path

from xeda import Design
from xeda.flow_runner import DefaultRunner
from xeda.flows import Quartus
from xeda.flows.flow import FPGA
from xeda.flows.quartus import parse_csv, try_num

TESTS_DIR = Path(__file__).parent.absolute()
RESOURCES_DIR = TESTS_DIR / "resources"
EXAMPLES_DIR = TESTS_DIR.parent / "examples"


log = logging.getLogger(__name__)

log.root.setLevel(logging.DEBUG)
log.setLevel(logging.DEBUG)


def test_parse_csv():
    resources = parse_csv(
        RESOURCES_DIR / "Fitter_Resource_Utilization_by_Entity.csv",
        id_field="Compilation Hierarchy Node",
        field_parser=lambda s: try_num(s.split()[0]),
        id_parser=lambda s: s.strip().lstrip("|"),
        # interesting_fields=None
        interesting_fields={
            "Logic Cells",
            "LUT-Only LCs",
            "Register-Only LCs",
            "LUT/Register LCs",
            "Dedicated Logic Registers",
            "ALMs needed [=A-B+C]",
            "Combinational ALUTs",
            "ALMs used for memory",
            "Memory Bits",
            "M10Ks",
            "M9Ks",
            "DSP Elements",
            "DSP Blocks",
            "Block Memory Bits",
            "Pins",
            "I/O Registers",
        }
        # ['Logic Cells', 'Memory Bits', 'M10Ks', 'M9Ks', 'DSP Elements', 'ALMs needed [=A-B+C]',
        #                     'Combinational ALUTs', 'ALMs used for memory', 'DSP Blocks', 'Pins'
        #                     'LUT-Only LCs',	'Register-Only LCs', 'LUT/Register LCs', 'Block Memory Bits']
    )
    assert resources == {
        "full_adder_piped": {
            "ALMs needed [=A-B+C]": 1.5,
            "ALMs used for memory": 0.0,
            "Block Memory Bits": 0,
            "Combinational ALUTs": 3,
            # 'Compilation Hierarchy Node': '|full_adder_piped',
            "DSP Blocks": 0,
            "Dedicated Logic Registers": 2,
            # 'Entity Name': 'full_adder_piped',
            # 'Full Hierarchy Name': '|full_adder_piped',
            "I/O Registers": 0,
            # 'Library Name': 'work',
            "M10Ks": 0,
            "Pins": 7,
            # 'Virtual Pins': 0,
            # '[A] ALMs used in final placement': 1.5,
            # '[B] Estimate of ALMs recoverable by dense packing': 0.0,
            # '[C] Estimate of ALMs unavailable': 0.0
        },
    }


def _test_parse_reports():
    root_dir = Path("/Users/kamyar/src/xeda/examples/vhdl/pipeline")
    design = Design.from_toml(root_dir / "pipelined_adder.toml")
    run_path = root_dir / "xeda_run/pipelined_adder/quartus"
    settings = Quartus.Settings(fpga={"part": "10CL016YU256C6G"}, clock_period=15)
    flow = Quartus(settings, design, run_path=run_path)
    flow.init()
    flow.parse_reports()
    print(flow.results)


def test_parse_csv_no_header():
    parsed = parse_csv(RESOURCES_DIR / "Flow_Summary.csv", None)
    expected = {
        "Flow Status": "Successful - Tue Mar  1 11:10:35 2022",
        "Quartus Prime Version": "21.1.0 Build 842 10/21/2021 SJ Lite Edition",
        "Revision Name": "pipelined_adder",
        "Top-level Entity Name": "full_adder_piped",
        "Family": "Cyclone V",
        "Device": "5CGXBC3B6F23C7",
        "Timing Models": "Final",
        "Total registers": "2",
        "Total pins": "7 / 222 ( 3 % )",
        "Total virtual pins": "0",
        "Total DSP Blocks": "0 / 57 ( 0 % )",
        "Total HSSI RX PCSs": "0 / 3 ( 0 % )",
        "Total HSSI PMA RX Deserializers": "0 / 3 ( 0 % )",
        "Total HSSI TX PCSs": "0 / 3 ( 0 % )",
        "Total HSSI PMA TX Serializers": "0 / 3 ( 0 % )",
        "Total PLLs": "0 / 7 ( 0 % )",
        "Total DLLs": "0 / 3 ( 0 % )",
    }

    for k, v in expected.items():
        assert parsed[k] == v


def prepend_to_path(path):
    current_path = os.environ.get("PATH", "").split(os.pathsep)
    current_path.insert(0, str(path))
    os.environ["PATH"] = os.pathsep.join(current_path)


def test_quartus_synth_py() -> None:
    path = RESOURCES_DIR / "design0/design0.toml"
    # Append to PATH so if the actual tool exists, would take precedences.
    dockerized = True
    if "PYTEST_CURRENT_TEST" in os.environ:
        prepend_to_path(TESTS_DIR / "fake_tools")
        dockerized = False
    assert path.exists()
    design = Design.from_toml(EXAMPLES_DIR / "vhdl" / "sqrt" / "sqrt.toml")
    settings = dict(fpga=FPGA("10CL016YU256C6G"), clock_period=6, dockerized=dockerized)
    with tempfile.TemporaryDirectory() as run_dir:
        print("Xeda run dir: ", run_dir)
        xeda_runner = DefaultRunner(run_dir, debug=True)
        flow = xeda_runner.run_flow(Quartus, design, settings)
        settings_json = flow.run_path / "settings.json"
        results_json = flow.run_path / "results.json"
        assert settings_json.exists()
        assert results_json.exists()
        assert flow.succeeded


if __name__ == "__main__":
    _test_parse_reports()
    # test_quartus_synth_py()



import logging
from typing import List, Optional

from ...dataclass import Field
from ...utils import SDF
from ..flow import SimFlow
from ..vivado import Vivado

log = logging.getLogger(__name__)


# FIXME: Does not return error when simulation is finished with a failure assertion


class VivadoSim(Vivado, SimFlow):
    """Simulate using Xilinx Vivado simulator (xsim) flow"""

    # TODO change this?
    # Can run multiple configurations (a.k.a testvectors) in a single run of Vivado through "run_configs"

    class Settings(Vivado.Settings, SimFlow.Settings):
        saif: Optional[str] = None
        elab_flags: List[str] = ["-relax"]
        analyze_flags: List[str] = ["-relax"]
        sim_flags: List[str] = []
        elab_debug: Optional[str] = None  # TODO choices: "typical", ...
        sdf: SDF = SDF()
        optimization_flags: List[str] = []  # ["-O3"]
        debug_traces: bool = False
        prerun_time: Optional[str] = None
        work_lib: str = "work"
        initialize_zeros: bool = Field(
            False, description="Initialize all signals with zero"
        )
        xelab_log: Optional[str] = "xeda_xelab.log"

    def run(self) -> None:
        ss = self.settings
        assert isinstance(ss, self.Settings)

        ss.elab_flags.append(f'-mt {"off" if ss.debug else ss.nthreads}')
        elab_debug = ss.elab_debug
        if not elab_debug and (ss.debug or ss.saif or ss.vcd):
            elab_debug = "typical"
        if elab_debug:
            ss.elab_flags.append(f"-debug {elab_debug}")

        assert self.design.tb
        if ss.vcd:
            log.info("Dumping VCD to %s", self.run_path / ss.vcd)
        sdf_root = ss.sdf.root
        if not sdf_root:
            sdf_root = self.design.tb.uut
        for delay_type, sdf_file in ss.sdf.delay_items():
            assert sdf_root, "neither SDF root nor tb.uut are provided"
            ss.elab_flags.append(f"-sdf{delay_type} {sdf_root}={sdf_file}")

        script_path = self.copy_from_template("vivado_sim.tcl")
        self.vivado.run("-source", script_path)



import logging
from typing import List, Literal, Optional
from ..flow import FpgaSynthFlow
from ...tool import Tool
from ...dataclass import validator

log = logging.getLogger(__name__)


class DiamondSynth(FpgaSynthFlow):
    class Settings(FpgaSynthFlow.Settings):
        impl_folder: str = "diamond_impl"
        impl_name: str = "Implementation0"
        syn_cmdline_args: Optional[List[str]] = None
        synthesis_engine: Literal["lse", "synplify"] = "lse"

        @validator("syn_cmdline_args", pre=True)
        def validate_syn_cmdline_args(
            cls, syn_cmdline_args: Optional[List[str]]
        ) -> List[str]:
            if syn_cmdline_args is None:
                syn_cmdline_args = []
            return syn_cmdline_args

    def run(self) -> None:
        assert isinstance(self.settings, self.Settings)
        constraint_exts = (
            ["ldc"] if self.settings.synthesis_engine == "lse" else ["sdc", "fdc"]
        )
        constraints = [f"constraints.{ext}" for ext in constraint_exts]
        for constraint in constraints:
            self.copy_from_template(constraint)
        script_path = self.copy_from_template("synth.tcl")
        diamondc = Tool("diamondc")
        diamondc.run("diamondc", script_path)

    def parse_reports(self) -> bool:
        assert isinstance(self.settings, self.Settings)
        reports_dir = self.run_path / "diamond_impl"
        design_name = self.design.name
        impl_name = self.settings.impl_name

        period_pat = r"""^\s*Preference:\s+PERIOD\s+PORT\s+\"(?P<clock_port>\w+)\"\s+(?P<clock_period>\d+\.\d+)\s+ns.*HIGH\s+\d+\.\d+\s+ns\s*;\s*
\s*\d+\s+items\s+\S+\s+(?P<_timing_errors>\d+)\s+timing\s+errors?"""
        freq_pat = r"^\s*Preference:\s+FREQUENCY\s+PORT\s+\"(?P<clock_port>\w+)\"\s+(?P<clock_frequency>\d+\.\d+)\s+MHz\s*;\s*\n\s*\d+\s+items\s+\S+\s+(?P<_timing_errors>\d+)\s+timing\s+errors?"
        self.parse_report_regex(
            reports_dir / f"{design_name}_{impl_name}.twr", [period_pat, freq_pat]
        )

        if "clock_frequency" in self.results:
            frequency = self.results["clock_frequency"]
            period = 1000.0 / frequency
            self.results["clock_period"] = period

        else:
            period = self.results["clock_period"]
            frequency = 1000.0 / period
            self.results["clock_frequency"] = frequency

        slice_pat = r"^Device\s+utilization\s+summary:\s*.*^\s+SLICE\s+(?P<slice>\d+)\/(?P<_slice_avail>\d+).*^Number\s+of\s+Signals"
        time_pat = r"""Level\s*/\s+Number\s+Worst\s+Timing\s+Worst\s+Timing\s+Run\s+NCD\s*
\s*Cost\s+\[ncd\]\s+Unrouted\s+Slack\s+Score\s+Slack\(hold\)\s+Score\(hold\)\s+Time\s+Status\s*
(\s*\-+){8}\s*
\s*(?P<_lvl_cost>\S+)\s+(?P<_ncd>\S+)\s+(?P<_num_unrouted>\d+)\s+(?P<wns>\-?\d+\.\d+)\s+(?P<_setup_score>\d+)\s+(?P<whs>\-?\d+\.\d+)\s+(?P<_hold_score>\d+)\s+(?P<_runtime>\d+(?:\:\d+)*)\s+(?P<_status>\w+)\s*$"""
        self.parse_report_regex(
            reports_dir / f"{design_name}_{impl_name}.par", slice_pat, time_pat
        )

        # NOTE there can be "page breaks" anywhere in the mrp file (others? TODO)
        # NOTE therefore only match lines
        #   1. Total number of LUT4s = (Number of logic LUT4s) + 2*(Number of distributed RAMs) + 2*(Number of ripple logic)
        #   2. Number of logic LUT4s does not include count of distributed RAM and ripple logic.
        slice_pattern = r"""\s*Number\s+of\s+registers:\s*(?P<ff>\d+)\s+out\s+of\s*(?P<_ff_avail>\d+).*
\s*Number\s+of\s+SLICEs:\s*(?P<_slice_map>\d+)\s*out\s+of\s*(?P<_slice_avail>\d+).*
\s+SLICEs\s+as\s+RAM:\s*(?P<_slice_ram>\d+)\s*out\s+of\s*(?P<_slice_ram_avail>\d+).*
\s+SLICEs\s+as\s+Carry:\s*(?P<_slice_carry>\d+)\s+out\s+of\s+(?P<_slice_carry_avail>\d+).*
\s*Number\s+of\s+LUT4s:\s*(?P<lut>\d+)\s+out of\s+(?P<_lut_avail>\d+).*
\s+Number\s+used\s+as\s+logic\s+LUTs:\s*(?P<_lut_logic>\d+)\s*
\s+Number\s+used\s+as\s+distributed\s+RAM:\s*(?P<_lut_dram>\d+)\s*
\s+Number\s+used\s+as\s+ripple\s+logic:\s*(?P<_lut_ripple>\d+)\s*
\s+Number\s+used\s+as\s+shift\s+registers:\s*(?P<_lut_shift>\d+)\s*.*
\s*Number\s+of\s+block\s+RAMs:\s*(?P<bram>\d+)\s+out\s+of\s+(?P<_bram_avail>\d+).*"""

        dsp_pattern = r"""\s+MULT18X18D\s+(?P<_dsp_MULT18X18D>\d+)\s*.*
\s+MULT9X9D\s+(?P<_dsp_MULT9X9D>\d+)\s*.*"""

        self.parse_report_regex(
            reports_dir / f"{design_name}_{impl_name}.mrp", slice_pattern, dsp_pattern
        )

        # FIXME add other types of available ALUs and DSPs
        self.results["dsp"] = (
            self.results["_dsp_MULT18X18D"] + self.results["_dsp_MULT9X9D"]
        )

        failed = False

        # TODO FIXME move to LwcSynth
        forbidden_resources = ["dsp", "bram"]
        for res in forbidden_resources:
            if res in self.results and self.results[res] != 0:
                log.critical(
                    "Map report shows %s use(s) of forbidden resource %s.",
                    self.results[res],
                    res,
                )
                failed = True

        failed = (
            failed
            or (self.results["wns"] < 0)
            or (self.results["whs"] < 0)
            or (self.results["_num_unrouted"] != 0)
            or (self.results["_status"].lower() != "completed")
            or (self.results["_timing_errors"] != 0)
        )
        return not failed
